@@ -1,22 +1,23 @@
 /* The big picture:
-   For every outgoing request, we have to have the network-ready data
-   somewhere in memory.
-
-   Using writev, any pieces that do not need endian conversion can
-   be written in-place.
-
-   The pieces that do need endian conversion can be put into one or more
-   buffers.
-
-   WHOA WHOA newsflash
-   Because IIOP lets the message sender specify the endianness,
-   we do not need to do endian conversion _ever_! The receiver can do all
-   conversions if need be, or if they are the same endianness as sender they
-   can just pull it in right off the wire :)
-   
+ * For every outgoing request, we have to have the network-ready data
+ * somewhere in memory.
+ *
+ * Using writev, any pieces that do not need endian conversion can
+ * be written in-place.
+ *
+ * The pieces that do need endian conversion can be put into one or more
+ * buffers.
+ *
+ * WHOA WHOA newsflash
+ * Because IIOP lets the message sender specify the endianness,
+ * we do not need to do endian conversion _ever_! The receiver can do all
+ * conversions if need be, or if they are the same endianness as sender they
+ * can just pull it in right off the wire :)
+ * 
  */
 
 #include "config.h"
+#include "iiop-endianP.h"
 #include <string.h>
 #include <unistd.h>
 #include <stdio.h>
@@ -32,29 +33,30 @@
 #include "IIOP-private.h"
 
 #ifdef HAVE_LIMITED_WRITEV
-#define writev proxy_writev
+#define writev g_writev
 #endif
 
 /* type defs */
-/* Overlaps with struct _GIOPMessageHeader on purpose
-   - we save time because this stuff never changes */
+
+/*
+ * Overlaps with struct _GIOPMessageHeader on purpose
+ *  - we save time because this stuff never changes
+ */
 struct _GIOPMessageHeaderConstants {
   GIOP_char magic[4];
-  GIOPVersion GIOP_version;
+  GIOP_char GIOP_version[2];
   GIOP_octet flags;
 };
 
 /* functions */
 static gint giop_recv_decode_message(GIOPRecvBuffer *buf);
-#if 0
 static gboolean num_on_list(GIOP_unsigned_long num,
 			    const GIOP_unsigned_long *request_ids,
 			    GIOP_unsigned_long req_count);
-#endif
 static gint giop_recv_reply_decode_message(GIOPRecvBuffer *buf);
 static gint giop_recv_request_decode_message(GIOPRecvBuffer *buf);
-static inline size_t giop_message_buffer_do_alignment(GIOPMessageBuffer *buffer,
-						      gulong align_for);
+static gint giop_recv_locate_reply_decode_message(GIOPRecvBuffer *buf);
+static gint giop_recv_locate_request_decode_message(GIOPRecvBuffer *buf);
 static GIOPRecvBuffer *giop_received_list_check_reply(GIOP_unsigned_long request_id);
 
 #ifdef _REENTRANT
@@ -77,10 +79,10 @@ giop_first_message_vec = {NULL,
 			  sizeof(struct _GIOPMessageHeaderConstants)};
 
 DEFINE_LOCK(sendbufferlist);
-GList *sendbufferlist = NULL;
+GSList *sendbufferlist = NULL;
 
 DEFINE_LOCK(recvbufferlist);
-GList *recvbufferlist = NULL;
+GSList *recvbufferlist = NULL;
 
 DEFINE_LOCK(incoming_bufs);
 GList *incoming_bufs = NULL; /* List of incoming messages that had to be
@@ -98,7 +100,7 @@ void giop_message_buffer_append_iovec(GIOPMessageBuffer *msgbuf,
 				      const struct iovec *iovec)
 {
   /* g_print("Appending iovec %d bytes @ %p\n", iovec->iov_len, iovec->iov_base); */
-  g_array_append_val(msgbuf->iovecs, struct iovec, *iovec);
+  g_array_append_val(msgbuf->iovecs, *iovec);
 }
 
 void
@@ -119,7 +121,7 @@ giop_message_buffer_init(void)
 static void
 giop_message_buffer_new(GIOPMessageBuffer *buf)
 {
-  buf->iovecs = g_array_new(FALSE);
+  buf->iovecs = g_array_new(FALSE, FALSE, sizeof(struct iovec));
 }
 
 #define STRUCT_OFFSET(t, f) ((int) ((char*) &((t*) 0)->f))
@@ -145,9 +147,8 @@ giop_send_buffer_new(void)
     - STRUCT_OFFSET(GIOPMessageHeader, message_type);
   GIOP_MESSAGE_BUFFER(msgbuf)->message_header.message_size = 0;
 
-  msgbuf->indirects = g_mem_chunk_new("GIOPSendBuffer:indirects",
-				      GIOP_INDIRECT_CHUNK_SIZE,
-				      2, G_ALLOC_ONLY);
+  msgbuf->indirects = g_mem_chunk_create(char[GIOP_INDIRECT_CHUNK_SIZE],
+					 2, G_ALLOC_ONLY);
 
   giop_message_buffer_append_iovec(GIOP_MESSAGE_BUFFER(msgbuf), &firstvec);
 
@@ -160,21 +161,42 @@ giop_send_buffer_write(GIOPSendBuffer *send_buffer)
   gulong nvecs;
   gulong res, i, sum;
 
-  nvecs = GIOP_MESSAGE_BUFFER(send_buffer)->iovecs->len / sizeof(struct iovec);
+#if 0
+  switch(GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_type) {
+  case GIOP_REQUEST:
+    g_message("[%d] sending request size %d on FD %d",
+	      getpid(),
+	      GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size,
+	      GIOP_CONNECTION_GET_FD(GIOP_MESSAGE_BUFFER(send_buffer)->connection));
+    break;
+  case GIOP_REPLY:
+    g_message("[%d] sending reply size %d on FD %d",
+	      getpid(),
+	      GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size,
+	      GIOP_CONNECTION_GET_FD(GIOP_MESSAGE_BUFFER(send_buffer)->connection));
+    break;
+  default:
+  }
+#endif
+
+  nvecs = GIOP_MESSAGE_BUFFER(send_buffer)->iovecs->len;
 
   res = 
     writev(GIOP_CONNECTION_GET_FD(GIOP_MESSAGE_BUFFER(send_buffer)->connection),
 	   (struct iovec *)GIOP_MESSAGE_BUFFER(send_buffer)->iovecs->data,
 	   nvecs);
 
-#ifdef DEBUG
   for(i = sum = 0; i < nvecs; i++) {
     sum += g_array_index((GIOP_MESSAGE_BUFFER(send_buffer)->iovecs), struct iovec, i).iov_len;
   }
 
+#ifdef DEBUG
   g_assert(sum == (GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size + sizeof(GIOPMessageHeader)));
-  g_assert(sum == res);
 #endif
+
+  if(sum != res) { /* Connection error */
+    giop_main_handle_connection_exception(GIOP_MESSAGE_BUFFER(send_buffer)->connection);
+  }
 }
 
 static GIOPSendBuffer *
@@ -188,9 +210,15 @@ giop_send_buffer_use(GIOPConnection *connection)
 
   if(sendbufferlist)
     {
+      GSList *head;
+
       retval = sendbufferlist->data;
-      sendbufferlist = g_list_remove_link(sendbufferlist, sendbufferlist);
-      g_array_truncate(GIOP_MESSAGE_BUFFER(retval)->iovecs, struct iovec, 2);
+
+      head = sendbufferlist;
+      sendbufferlist = g_slist_remove_link(sendbufferlist, sendbufferlist);
+      g_slist_free_1 (head);
+
+      g_array_set_size(GIOP_MESSAGE_BUFFER(retval)->iovecs, 2);
       GIOP_MESSAGE_BUFFER(retval)->message_header.message_size = 0;
     }
   else
@@ -202,7 +230,10 @@ giop_send_buffer_use(GIOPConnection *connection)
   GIOP_MESSAGE_BUFFER(retval)->connection = connection;
 
   g_mem_chunk_reset(retval->indirects);
-  retval->indirect = g_mem_chunk_alloc(retval->indirects);
+  retval->indirect = g_chunk_new(gpointer, retval->indirects);
+#ifdef DEBUG
+  memset(retval->indirect, '\xFE', GIOP_INDIRECT_CHUNK_SIZE);
+#endif
   retval->indirect_used = 0;
 
   return retval;
@@ -227,6 +258,23 @@ giop_send_reply_buffer_use(GIOPConnection *connection,
   return send_buffer;
 }
 
+GIOPSendBuffer *
+giop_send_locate_reply_buffer_use(GIOPConnection *connection,
+			   GIOP_unsigned_long request_id,
+			   GIOPLocateStatusType locate_reply_status)
+{
+	GIOPSendBuffer *send_buffer;
+
+	send_buffer = giop_send_buffer_use(connection);
+
+	GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_type = GIOP_LOCATEREPLY;
+
+	APIA(&request_id, sizeof(request_id));
+	APIA(&locate_reply_status, sizeof(locate_reply_status));
+
+	return send_buffer;
+}
+
 GIOPSendBuffer *
 giop_send_request_buffer_use(GIOPConnection *connection,
 			     const IOP_ServiceContextList *service_context,
@@ -234,7 +282,7 @@ giop_send_request_buffer_use(GIOPConnection *connection,
 			     GIOP_boolean response_expected,
 			     const struct iovec *object_key_vec,
 			     const struct iovec *operation_vec,
-			     CORBA_Principal *requesting_principal)
+			     const struct iovec *principal_vec)
 {
   GIOPSendBuffer *send_buffer;
 #if 0
@@ -247,6 +295,16 @@ giop_send_request_buffer_use(GIOPConnection *connection,
    sizeof(CORBA_unsigned_long) + sizeof("nobody")};
 #endif
 
+  if (!connection)
+	  return NULL;
+  g_return_val_if_fail(object_key_vec, NULL);
+  g_return_val_if_fail(operation_vec, NULL);
+
+ ORBit_Trace(TraceMod_IIOP, TraceLevel_Debug,
+             "Sending request %s id %d to %s\n",
+	     operation_vec->iov_base + 4,
+	     request_id, object_key_vec->iov_base + 4);
+
   send_buffer = giop_send_buffer_use(connection);
 
   GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_type = GIOP_REQUEST;
@@ -265,8 +323,10 @@ giop_send_request_buffer_use(GIOPConnection *connection,
   GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size +=
     object_key_vec->iov_len;
 
-  /* We can know the length at compile time - don't calculate it at runtime
-     if we can help it :) */
+  /*
+   * We can know the length at compile time - don't calculate it at runtime
+   * if we can help it :)
+   */
   /* ENCODER_CALL(CORBA_string, (CORBA_string *)operation); */
   giop_message_buffer_do_alignment(GIOP_MESSAGE_BUFFER(send_buffer),
 				   sizeof(CORBA_unsigned_long));
@@ -275,22 +335,48 @@ giop_send_request_buffer_use(GIOPConnection *connection,
   GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size +=
     operation_vec->iov_len;
 
-    ENCODER_CALL(CORBA_Principal, requesting_principal);
-#if 0
-  if(requesting_principal)
-  else {
-    giop_message_buffer_do_alignment(GIOP_MESSAGE_BUFFER(send_buffer),
-				     sizeof(CORBA_unsigned_long));
-    giop_message_buffer_append_iovec(GIOP_MESSAGE_BUFFER(send_buffer),
-				     &default_principal_vec);
-    GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size +=
-      default_principal_vec.iov_len;
-  }
-#endif
+  giop_message_buffer_do_alignment(GIOP_MESSAGE_BUFFER(send_buffer),
+				   sizeof(CORBA_unsigned_long));
+  giop_message_buffer_append_iovec(GIOP_MESSAGE_BUFFER(send_buffer),
+				   principal_vec);
+  GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size +=
+    principal_vec->iov_len;
 
   return send_buffer;
 }
 
+GIOPSendBuffer *
+giop_send_locate_request_buffer_use(GIOPConnection *connection,
+			     GIOP_unsigned_long request_id,
+			     const struct iovec *object_key_vec)
+{
+	GIOPSendBuffer *send_buffer;
+
+	if (!connection)
+		return NULL;
+
+	g_return_val_if_fail(object_key_vec, NULL);
+
+	ORBit_Trace(TraceMod_IIOP, TraceLevel_Debug,
+		"Sending locate request id %d to %s\n",
+		request_id, object_key_vec->iov_base + 4);
+
+	send_buffer = giop_send_buffer_use(connection);
+
+	GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_type = GIOP_LOCATEREQUEST;
+
+	APIA(&request_id, sizeof(request_id));
+
+	giop_message_buffer_do_alignment(GIOP_MESSAGE_BUFFER(send_buffer),
+		sizeof(CORBA_unsigned_long));
+	giop_message_buffer_append_iovec(GIOP_MESSAGE_BUFFER(send_buffer),
+		object_key_vec);
+	GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size +=
+		object_key_vec->iov_len;
+
+	return send_buffer;
+}
+
 void
 giop_send_buffer_unuse(GIOPSendBuffer *send_buffer)
 {
@@ -299,11 +385,11 @@ giop_send_buffer_unuse(GIOPSendBuffer *send_buffer)
   giop_connection_unref(GIOP_MESSAGE_BUFFER(send_buffer)->connection);
 
   GET_LOCK(sendbufferlist);
-  sendbufferlist = g_list_prepend(sendbufferlist, send_buffer);
+  sendbufferlist = g_slist_prepend(sendbufferlist, send_buffer);
   RELEASE_LOCK(sendbufferlist);
 }
 
-static inline size_t
+gulong
 giop_message_buffer_do_alignment(GIOPMessageBuffer *buffer,
 				 gulong align_for)
 {
@@ -312,15 +398,16 @@ giop_message_buffer_do_alignment(GIOPMessageBuffer *buffer,
   guint alignme;
 
   if(align_for < 2) return 0;
-  if(align_for > sizeof(long)) align_for = sizeof(long);
+  if(align_for >
+     MAX(sizeof(GIOP_long_long),sizeof(GIOP_long_double)))
+    align_for = MAX(sizeof(GIOP_long_long), sizeof(GIOP_long_double));
 
   alignme = (gulong)ALIGN_ADDRESS(buffer->message_header.message_size, align_for);
 
   if((alignme - buffer->message_header.message_size) > 0)
     {
-      lastvec = (struct iovec *)(buffer->iovecs->data
-				 + buffer->iovecs->len
-				 - sizeof(struct iovec));
+      lastvec = (struct iovec *)(buffer->iovecs->data) 
+				 + buffer->iovecs->len - 1;
 
       if(lastvec->iov_base == giop_scratch_space)
 	{
@@ -354,9 +441,8 @@ giop_message_buffer_append_mem_a(GIOPMessageBuffer *buffer,
 
   alignfor = giop_message_buffer_do_alignment(buffer, mem_region_length);
 
-  lastvec = (struct iovec *)(buffer->iovecs->data
-			     + buffer->iovecs->len
-			     - sizeof(struct iovec));
+  lastvec = (struct iovec *)(buffer->iovecs->data) + 
+			     + buffer->iovecs->len - 1;
 
   if((mem_region == giop_scratch_space && lastvec->iov_base == giop_scratch_space)
      || (alignfor == 0 && (lastvec->iov_base + lastvec->iov_len) == mem_region))
@@ -381,13 +467,12 @@ giop_message_buffer_append_mem(GIOPMessageBuffer *buffer,
   struct iovec newvec;
   struct iovec *lastvec;
 
-  lastvec = (struct iovec *)(buffer->iovecs->data
-			     + buffer->iovecs->len
-			     - sizeof(struct iovec));
+  lastvec = (struct iovec *)(buffer->iovecs->data)
+			     + buffer->iovecs->len - 1;
 
   if((mem_region == giop_scratch_space
       && lastvec->iov_base == giop_scratch_space)
-     || ((lastvec->iov_base + lastvec->iov_len + 1) == mem_region))
+     || ((lastvec->iov_base + lastvec->iov_len) == mem_region))
     {
       lastvec->iov_len += mem_region_length;
     }
@@ -401,46 +486,61 @@ giop_message_buffer_append_mem(GIOPMessageBuffer *buffer,
   buffer->message_header.message_size += mem_region_length;
 }
 
-void
+gpointer
 giop_send_buffer_append_mem_indirect(GIOPSendBuffer *send_buffer,
 				     gconstpointer mem_region,
 				     gulong mem_region_length)
 {
+  gpointer blockstart = NULL;
   gulong offset, new_offset;
-  
+
   for(offset = new_offset = 0; new_offset < mem_region_length;)
     {
       new_offset =
 	MIN(mem_region_length - offset,
 	    GIOP_INDIRECT_CHUNK_SIZE - send_buffer->indirect_used);
 
-      memcpy(send_buffer->indirect + send_buffer->indirect_used,
-	     mem_region + offset, new_offset - offset);
+      if((new_offset - offset) > sizeof(CORBA_unsigned_long)
+	 || mem_region_length >= sizeof(CORBA_unsigned_long)) {
 
-      giop_message_buffer_append_mem(GIOP_MESSAGE_BUFFER(send_buffer),
-				     send_buffer->indirect + offset,
-				     new_offset - offset);
+	if(!blockstart)
+	  blockstart = send_buffer->indirect + send_buffer->indirect_used;
+      }
+	
+	memcpy(send_buffer->indirect + send_buffer->indirect_used,
+	       mem_region + offset, new_offset - offset);
+	
+	giop_message_buffer_append_mem(GIOP_MESSAGE_BUFFER(send_buffer),
+				       send_buffer->indirect + send_buffer->indirect_used,
+				       new_offset - offset);
+
+	send_buffer->indirect_used += new_offset - offset;
 
       offset = new_offset;
 
-      if(new_offset == mem_region_length)
+      if(new_offset >= mem_region_length)
 	{
 	  send_buffer->indirect_used = 0;
 	  send_buffer->indirect = g_chunk_new(gpointer,
-						 send_buffer->indirects);
+					      send_buffer->indirects);
+#ifdef DEBUG
+	  memset(send_buffer->indirect, '\xFE', GIOP_INDIRECT_CHUNK_SIZE);
+#endif
 	}
     }
+
+  return blockstart;
 }
 
-void
+gpointer
 giop_send_buffer_append_mem_indirect_a(GIOPSendBuffer *send_buffer,
 				       gconstpointer mem_region,
 				       gulong mem_region_length)
 {
   giop_message_buffer_do_alignment(GIOP_MESSAGE_BUFFER(send_buffer),
 				   mem_region_length);
-  giop_send_buffer_append_mem_indirect(send_buffer,
-				       mem_region, mem_region_length);
+  return giop_send_buffer_append_mem_indirect(send_buffer,
+					      mem_region, mem_region_length);
 }
 
 GIOP_unsigned_long
@@ -467,6 +567,7 @@ giop_recv_buffer_new(void)
   RELEASE_LOCK(recvbuffers);
 
   giop_message_buffer_new(GIOP_MESSAGE_BUFFER(msgbuf));
+  msgbuf->message_body = NULL;
 
   return msgbuf;
 }
@@ -479,7 +580,7 @@ giop_recv_buffer_unuse(GIOPRecvBuffer *buffer)
   g_free(buffer->message_body);
   giop_connection_unref(GIOP_MESSAGE_BUFFER(buffer)->connection);
   GET_LOCK(recvbufferlist);
-  recvbufferlist = g_list_prepend(recvbufferlist, buffer);
+  recvbufferlist = g_slist_prepend(recvbufferlist, buffer);
   RELEASE_LOCK(recvbufferlist);
 }
 
@@ -494,9 +595,16 @@ giop_recv_buffer_use(GIOPConnection *connection)
 
   if(recvbufferlist)
     {
+      GSList *head;
+
       retval = recvbufferlist->data;
-      recvbufferlist = g_list_remove_link(recvbufferlist, recvbufferlist);
+
+      head = recvbufferlist;
+      recvbufferlist = g_slist_remove_link(recvbufferlist, recvbufferlist);
+      g_slist_free_1 (head);
+      
       GIOP_MESSAGE_BUFFER(retval)->message_header.message_size = 0;
+      retval->message_body = NULL;
     }
   else
     retval = giop_recv_buffer_new();
@@ -514,33 +622,49 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
 {
   GIOPRecvBuffer *retval;
   int sysret;
+  int bytes_read;
+  int n_zero_reads;
+  int message_size;
+  
+  if (!connection)
+	  return NULL;
 
-  g_return_val_if_fail(connection != NULL, NULL);
   g_return_val_if_fail(connection->is_valid, NULL);
 
   retval = giop_recv_buffer_use(connection);
 
   if(!retval) return retval;
 
-  sysret = read(GIOP_CONNECTION_GET_FD(connection),
-		&(GIOP_MESSAGE_BUFFER(retval)->message_header),
-		sizeof(GIOP_MESSAGE_BUFFER(retval)->message_header));
+  bytes_read = n_zero_reads = 0;
+  do
+    {
+      sysret = read(GIOP_CONNECTION_GET_FD(connection),
+		    ((char *)&(GIOP_MESSAGE_BUFFER(retval)->message_header))+bytes_read,
+		    sizeof(GIOP_MESSAGE_BUFFER(retval)->message_header)-bytes_read);
 
-  if(sysret <= 0) {
-    giop_recv_buffer_unuse(retval);
-    g_warning("read() header [%d] failed = %d: %s",
-	      sizeof(GIOP_MESSAGE_BUFFER(retval)->message_header),
-	      sysret,
-	      g_strerror(errno));
+      if(sysret <= 0) {
+	giop_main_handle_connection_exception(connection);
+	giop_recv_buffer_unuse(retval);
+	return NULL;
+      }
+
+      bytes_read += sysret;
+    } while( bytes_read != 
+	     sizeof(GIOP_MESSAGE_BUFFER(retval)->message_header) );
+
+  /* Check the magic stuff */
+  if(strncmp(GIOP_MESSAGE_BUFFER(retval)->message_header.magic, "GIOP", 4)
+     || GIOP_MESSAGE_BUFFER(retval)->message_header.GIOP_version[0] != 1) {
     giop_main_handle_connection_exception(connection);
+    giop_recv_buffer_unuse(retval);
     return NULL;
   }
 
   if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(retval)))
     {
       CORBA_unsigned_long t = GIOP_MESSAGE_BUFFER(retval)->message_header.message_size;
-      retval->decoder = (gpointer)byteswap;
-      byteswap((gpointer)&GIOP_MESSAGE_BUFFER(retval)->message_header.message_size,
+      retval->decoder = (gpointer)iiop_byteswap;
+      iiop_byteswap((gpointer)&GIOP_MESSAGE_BUFFER(retval)->message_header.message_size,
 	       (gpointer)&t, sizeof(t));
     }
   else
@@ -556,21 +680,42 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
     return NULL;
   }
 
-  /* Think up a more efficient way of doing this */
-  retval->message_body = g_malloc(GIOP_MESSAGE_BUFFER(retval)->message_header.message_size);
+  message_size = GIOP_MESSAGE_BUFFER(retval)->message_header.message_size;
 
-  sysret = read(GIOP_CONNECTION_GET_FD(connection), retval->message_body,
-		GIOP_MESSAGE_BUFFER(retval)->message_header.message_size);
-
-  if(sysret < 0) {
-    giop_recv_buffer_unuse(retval);
-    g_warning("read() body [%d] failed = %d: %s",
-	      GIOP_MESSAGE_BUFFER(retval)->message_header.message_size,
-	      sysret,
-	      g_strerror(errno));
-    giop_main_handle_connection_exception(connection);
+  /*
+   * We set a limit on the message size on an incoming connection
+   * to avoid a denial of service attack.
+   */
+  
+  if (message_size > 8192 || message_size < 0){
+    g_warning("initial message size is bigger than 8k (%d)", message_size);
+    giop_main_handle_connection_exception (connection);
     return NULL;
   }
+  
+  /* Think up a more efficient way of doing this */
+  retval->message_body = g_malloc(message_size);
+
+  bytes_read = 0;
+  n_zero_reads = 0;
+  do
+    {
+      sysret = read(GIOP_CONNECTION_GET_FD(connection), 
+		    retval->message_body + bytes_read,
+		    message_size - bytes_read); 
+
+      if(sysret < 0 || (sysret == 0 && (n_zero_reads++ > 5))) {
+	giop_recv_buffer_unuse(retval);
+	g_warning("read() body [%d] failed = %d: %s",
+		  message_size,
+		  sysret,
+		  g_strerror(errno));
+	giop_main_handle_connection_exception(connection);
+	return NULL;
+      }
+      bytes_read += sysret;
+    } while( bytes_read != 
+	     message_size);
 
   retval->cur = retval->message_body;
 
@@ -587,6 +732,7 @@ void
 giop_received_list_push(GIOPRecvBuffer *recv_buffer)
 {
   GET_LOCK(incoming_bufs);
+  g_message("[%d] Adding buffer %p onto received list", getpid(), recv_buffer);
   incoming_bufs = g_list_prepend(incoming_bufs, recv_buffer);
   RELEASE_LOCK(incoming_bufs);
 }
@@ -605,6 +751,7 @@ GIOPRecvBuffer *giop_received_list_pop(void)
 
   retval = head->data;
   incoming_bufs = g_list_remove_link(incoming_bufs, head);
+  g_list_free_1 (head);
 
   RELEASE_LOCK(incoming_bufs);
 
@@ -635,25 +782,103 @@ giop_received_list_check_reply(GIOP_unsigned_long request_id)
   return retval;
 }
 
+/** giop_recv_reply_buffer_use_multiple
+ */
+GIOPRecvBuffer *
+giop_recv_reply_buffer_use_multiple(GArray *request_ids,
+				    gboolean block_for_reply)
+{
+  return giop_recv_reply_buffer_use_multiple_2(NULL, request_ids, block_for_reply);
+}
+
+GIOPRecvBuffer *
+giop_recv_reply_buffer_use_multiple_2(GIOPConnection *request_cnx,
+				      GArray *request_ids,
+				      gboolean block_for_reply)
+{
+  int i;
+  GIOPRecvBuffer *retval = NULL;
+
+  do {
+    /*
+     * We _do_ want to put this inside the loop,
+     * because we may call ourselves recursively for different request_id's
+     */
+    for(i = 0; i < request_ids->len && !retval; i++)
+      retval = giop_received_list_check_reply(g_array_index(request_ids, GIOP_unsigned_long, i));
+
+    if(retval)
+      break;
+
+    retval = giop_main_next_message_2(TRUE, request_cnx);
+
+    if(retval) {
+      if(GIOP_MESSAGE_BUFFER(retval)->message_header.message_type == GIOP_REPLY) {
+	 if(num_on_list(retval->message.u.reply.request_id,
+			(GIOP_unsigned_long *)request_ids->data,
+			request_ids->len))
+	   break;
+	 else
+	  giop_received_list_push(retval);
+      } else {
+	if(IIOPIncomingMessageHandler)
+	  IIOPIncomingMessageHandler(retval);
+	else
+	  giop_received_list_push(retval);
+	retval = NULL;
+      }
+    } else
+      return NULL;
+
+  } while(!retval && block_for_reply);
+
+  g_message("[%d] returning buffer %p from recv_reply", getpid(), retval);
+  return retval;
+}
+
 GIOPRecvBuffer *
 giop_recv_reply_buffer_use(GIOP_unsigned_long request_id,
 			   gboolean block_for_reply)
+{
+  return giop_recv_reply_buffer_use_2(NULL, request_id, block_for_reply);
+}
+
+GIOPRecvBuffer *
+giop_recv_reply_buffer_use_2(GIOPConnection *request_cnx,
+			     GIOP_unsigned_long request_id,
+			     gboolean block_for_reply)
+{
+  GArray fakeme;
+
+  fakeme.len = 1;
+  fakeme.data = (gpointer)&request_id;
+
+  return giop_recv_reply_buffer_use_multiple_2(request_cnx,
+					       &fakeme,
+					       block_for_reply);
+}
+
+GIOPRecvBuffer *
+giop_recv_locate_reply_buffer_use(GIOP_unsigned_long request_id,
+				  gboolean block_for_reply)
 {
   GIOPRecvBuffer *retval = NULL;
 
   do {
-    /* We _do_ want to put this inside the loop,
-       because we may call ourselves recursively for different request_id's */
+    /*
+     * We _do_ want to put this inside the loop,
+     * because we may call ourselves recursively for different request_id's
+     */
     retval = giop_received_list_check_reply(request_id);
 
     if(retval)
       break;
 
-    retval = giop_main_next_message(TRUE);
+    retval = giop_main_next_message_2(TRUE, NULL);
 
     if(retval) {
-      if(GIOP_MESSAGE_BUFFER(retval)->message_header.message_type == GIOP_REPLY
-	 && retval->message.u.reply.request_id == request_id)
+      if(GIOP_MESSAGE_BUFFER(retval)->message_header.message_type == GIOP_LOCATEREPLY
+	 && retval->message.u.locate_reply.request_id == request_id)
 	break;
       else {
 	if(IIOPIncomingMessageHandler)
@@ -680,17 +905,23 @@ giop_recv_decode_message(GIOPRecvBuffer *buf)
     case GIOP_REQUEST:
       return giop_recv_request_decode_message(buf);
       break;
+    case GIOP_LOCATEREQUEST:
+      return(giop_recv_locate_request_decode_message(buf));
+      break;
+    case GIOP_LOCATEREPLY:
+      return(giop_recv_locate_reply_decode_message(buf));
+      break;
     case GIOP_CLOSECONNECTION:
       return 0;
       break;
     default:
-      g_warning("Don't know how to decode message type %d\n",
+      g_warning("Don't know how to decode message type %d",
 		GIOP_MESSAGE_BUFFER(buf)->message_header.message_type);
       return -1;
     }
 }
 
-# define CHECK_POS(buf) g_return_val_if_fail(GIOP_RECV_BUFFER(buf)->cur < (GIOP_RECV_BUFFER(buf)->message_body + GIOP_MESSAGE_BUFFER(buf)->message_header.message_size), -1)
+# define CHECK_POS(buf) g_return_val_if_fail(GIOP_RECV_BUFFER(buf)->cur <= (GIOP_RECV_BUFFER(buf)->message_body + GIOP_MESSAGE_BUFFER(buf)->message_header.message_size), -1)
 
 /* There be dragons in here. */
 static gint
@@ -717,7 +948,7 @@ giop_recv_reply_decode_message(GIOPRecvBuffer *buf)
   if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(buf))) 
     {
 
-#define GET_ULONG(x) ({ byteswap((guchar *)&(x), \
+#define GET_ULONG(x) ({ iiop_byteswap((guchar *)&(x), \
 			      (guchar *)((buf->cur)),\
 			       sizeof(CORBA_unsigned_long)); buf->cur +=sizeof(CORBA_unsigned_long); })
 
@@ -762,6 +993,42 @@ giop_recv_reply_decode_message(GIOPRecvBuffer *buf)
       GET_ULONG_NC(buf->message.u.reply.reply_status);
     }
 
+#undef GET_ULONG  
+#undef GET_ULONG_NC
+
+#if 0
+  g_message("[%d] Received reply %d size %d to request %d",
+	    getpid(),
+	    buf->message.u.reply.reply_status,
+	    GIOP_MESSAGE_BUFFER(buf)->message_header.message_size,
+	    buf->message.u.reply.request_id);
+#endif
+
+  return 0;
+}
+
+static gint
+giop_recv_locate_reply_decode_message(GIOPRecvBuffer *buf)
+{
+  CHECK_POS(buf);
+  if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(buf))) 
+    {
+
+#define GET_ULONG(x) ({ iiop_byteswap((guchar *)&(x), \
+			      (guchar *)((buf->cur)),\
+			       sizeof(CORBA_unsigned_long)); buf->cur +=sizeof(CORBA_unsigned_long); })
+
+      GET_ULONG(buf->message.u.locate_reply.request_id);
+      GET_ULONG(buf->message.u.locate_reply.locate_status);
+    }
+  else
+    {
+#define GET_ULONG_NC(x) ({ (x) = (*((CORBA_unsigned_long *)(buf->cur))); buf->cur+=sizeof(CORBA_unsigned_long); x; })
+
+      GET_ULONG_NC(buf->message.u.locate_reply.request_id);
+      GET_ULONG_NC(buf->message.u.locate_reply.locate_status);
+    }
+
 #undef GET_ULONG  
 #undef GET_ULONG_NC
 
@@ -778,7 +1045,7 @@ giop_recv_request_decode_message(GIOPRecvBuffer *buf)
   if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(buf))) 
     {
 
-#define GET_ULONG(x) ({ byteswap((guchar *)&(x), \
+#define GET_ULONG(x) ({ iiop_byteswap((guchar *)&(x), \
 			      (guchar *)((buf->cur)),\
 			       sizeof(CORBA_unsigned_long)); buf->cur +=sizeof(CORBA_unsigned_long); })
       CHECK_POS(buf);
@@ -858,13 +1125,60 @@ giop_recv_request_decode_message(GIOPRecvBuffer *buf)
       CHECK_POS(buf);
     }
 
+#undef GET_ULONG  
+#undef GET_ULONG_NC
+
+#if 0
+  g_message("[%d] Received request %s size %d ID %d",
+	    getpid(),
+	    buf->message.u.request.operation,
+	    GIOP_MESSAGE_BUFFER(buf)->message_header.message_size,
+	    buf->message.u.request.request_id);
+#endif
+
+  return 0;
+}
+
+static gint
+giop_recv_locate_request_decode_message(GIOPRecvBuffer *buf)
+{
+  if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(buf))) 
+    {
+
+#define GET_ULONG(x) ({ iiop_byteswap((guchar *)&(x), \
+			      (guchar *)((buf->cur)),\
+			       sizeof(CORBA_unsigned_long)); buf->cur +=sizeof(CORBA_unsigned_long); })
+      CHECK_POS(buf);
+      buf->cur = ALIGN_ADDRESS(buf->cur, sizeof(GIOP_unsigned_long));
+      GET_ULONG(buf->message.u.locate_request.request_id);
+      CHECK_POS(buf);
+      buf->cur = ALIGN_ADDRESS(buf->cur, sizeof(GIOP_unsigned_long));
+      GET_ULONG(buf->message.u.locate_request.object_key._length);
+      buf->message.u.locate_request.object_key._buffer = buf->cur;
+      buf->cur += buf->message.u.locate_request.object_key._length;
+      CHECK_POS(buf);
+    }
+  else
+    {
+#define GET_ULONG_NC(x) ({ (x) = (*((CORBA_unsigned_long *)(buf->cur))); buf->cur+=sizeof(CORBA_unsigned_long); x; })
+
+      CHECK_POS(buf);
+      buf->cur = ALIGN_ADDRESS(buf->cur, sizeof(GIOP_unsigned_long));
+      GET_ULONG_NC(buf->message.u.locate_request.request_id);
+      CHECK_POS(buf);
+      buf->cur = ALIGN_ADDRESS(buf->cur, sizeof(GIOP_unsigned_long));
+      GET_ULONG_NC(buf->message.u.locate_request.object_key._length);
+      buf->message.u.locate_request.object_key._buffer = buf->cur;
+      buf->cur += buf->message.u.locate_request.object_key._length;
+      CHECK_POS(buf);
+    }
+
 #undef GET_ULONG  
 #undef GET_ULONG_NC
 
   return 0;
 }
 
-#if 0
 gboolean
 num_on_list(GIOP_unsigned_long num,
 	    const GIOP_unsigned_long *request_ids,
@@ -879,7 +1193,6 @@ num_on_list(GIOP_unsigned_long num,
 
   return FALSE;
 }
-#endif
 
 gboolean giop_msg_conversion_needed(GIOPMessageBuffer *msgbuf)
 {
