@@ -35,7 +35,9 @@ static void cbe_hdr_do_type_dcl(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_const_dcl(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_except_dcl(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_attr_dcl(CBEHeaderInfo *hinfo);
+static void cbe_hdr_do_inherited_attr_dcl(CBEHeaderInfo *hinfo, IDL_tree current_interface);
 static void cbe_hdr_do_op_dcl(CBEHeaderInfo *hinfo);
+static void cbe_hdr_do_inherited_op_dcl(CBEHeaderInfo *hinfo, IDL_tree current_interface);
 static void cbe_hdr_do_param_dcl(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_forward_dcl(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_integer(CBEHeaderInfo *hinfo);
@@ -49,6 +51,7 @@ static void cbe_hdr_do_type_boolean(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_octet(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_any(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_object(CBEHeaderInfo *hinfo);
+static void cbe_hdr_do_type_typecode(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_enum(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_sequence(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_type_array(CBEHeaderInfo *hinfo);
@@ -60,6 +63,7 @@ static void cbe_hdr_do_interface(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_module(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_binop(CBEHeaderInfo *hinfo);
 static void cbe_hdr_do_unaryop(CBEHeaderInfo *hinfo);
+static void cbe_hdr_do_native(CBEHeaderInfo *hinfo);
 
 void
 orbit_cbe_write_header(FILE *outfile, IDL_ns ns, IDL_tree tree)
@@ -68,10 +72,24 @@ orbit_cbe_write_header(FILE *outfile, IDL_ns ns, IDL_tree tree)
 
   g_return_if_fail(IDL_NODE_TYPE(tree) == IDLN_LIST);
 
+  fprintf(outfile, "/*\n"
+  		   " * This file was generated by orbit-idl - DO NOT EDIT!\n"
+		   " */\n\n");
   fprintf(outfile, "#include <glib.h>\n");
+  fprintf(outfile, "#define ORBIT_IDL_SERIAL %d\n", ORBIT_SERIAL);
   fprintf(outfile, "#include <orb/orbit.h>\n\n");
 
+  fprintf(outfile, "#ifdef __cplusplus\n");
+  fprintf(outfile, "extern \"C\" {\n");
+  fprintf(outfile, "#endif /* __cplusplus */\n\n");
+
   orbit_cbe_hdr_process_piece(&hinfo);
+
+  fprintf(outfile, "#ifdef __cplusplus\n");
+  fprintf(outfile, "}\n");
+  fprintf(outfile, "#endif /* __cplusplus */\n\n");
+
+  fprintf(outfile, "#undef ORBIT_IDL_SERIAL\n");
 }
 
 static void
@@ -206,7 +224,18 @@ orbit_cbe_hdr_process_piece(CBEHeaderInfo *hinfo)
     cbe_hdr_do_wide_string(hinfo);
     break;
   case IDLN_ANY:
-    g_error("IDLN_ANY not handled!\n");
+    g_error("IDLN_ANY not handled!");
+    break;
+  case IDLN_NATIVE:
+    cbe_hdr_do_native(hinfo);
+    break;
+  case IDLN_TYPE_TYPECODE:
+    cbe_hdr_do_type_typecode(hinfo);
+    break;
+  default:
+    /* Should never get these */
+    g_error("orbit_cbe_hdr_process_piece: Unhandled node type %s (%d)",
+	    IDL_NODE_TYPE_NAME(hinfo->tree), IDL_NODE_TYPE(hinfo->tree));
     break;
   }
 }
@@ -219,20 +248,24 @@ cbe_hdr_do_module(CBEHeaderInfo *hinfo)
 
   id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_MODULE(hinfo->tree).ident), "_", 0);
   g_return_if_fail(id != NULL);
-  idupper = malloc(strlen(id));
+  idupper = malloc(strlen(id) + 1);
 
   for(i = 0; id[i]; i++)
     idupper[i] = toupper(id[i]);
   idupper[i] = '\0';
 
+#if 0
   fprintf(hinfo->of, "#ifndef __%s__\n#define __%s__ 1\n", idupper, idupper);
+#endif
   fprintf(hinfo->of, "/* Begin CORBA module %s */\n\n", id);
 
   hinfo->tree = IDL_MODULE(hinfo->tree).definition_list;
   cbe_hdr_do_list(hinfo);
   
   fprintf(hinfo->of, "/* End CORBA module %s */\n", id);  
-  fprintf(hinfo->of, "#endif /* __%s__ */\n", idupper);
+#if 0
+  fprintf(hinfo->of, "#endif /* __%s__ */\n\n", idupper);
+#endif
   free(id); free(idupper);
 }
 
@@ -311,7 +344,11 @@ cbe_hdr_do_boolean(CBEHeaderInfo *hinfo)
 static void
 cbe_hdr_do_ident(CBEHeaderInfo *hinfo)
 {
-  g_assert(!"Not yet implemented");
+  char *id;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(hinfo->tree), "_", 0);
+  fprintf(hinfo->of, "typedef void *%s;\n", id);
+  free(id);
 }
 
 static void
@@ -324,7 +361,26 @@ cbe_hdr_do_type_dcl(CBEHeaderInfo *hinfo)
   for(curitem = IDL_TYPE_DCL(hinfo->tree).dcls;
       curitem; curitem = IDL_LIST(curitem).next) {
 
-    fprintf(hinfo->of, "typedef ");
+    if(IDL_NODE_TYPE(IDL_LIST(curitem).data) == IDLN_IDENT) {
+      id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_LIST(curitem).data),
+				   "_", 0);
+      fprintf(hinfo->of, "#ifndef _%s_defined\n"
+			 "#define _%s_defined\n\n", id, id);
+      fprintf(hinfo->of, "typedef ");
+      free(id);
+    } else if(IDL_NODE_TYPE(IDL_LIST(curitem).data) == IDLN_TYPE_ARRAY) {
+      id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_ARRAY(IDL_LIST(curitem).data).ident), "_", 0);
+
+      /* This is messy - the corresponding endif gets written in
+       * cbe_hdr_do_type_array
+       */
+      fprintf(hinfo->of, "#ifndef _%s_defined\n"
+			 "#define _%s_defined\n\n", id, id);
+      fprintf(hinfo->of, "typedef ");
+      free(id);
+    } else {
+      fprintf(hinfo->of, "/* not IDLN_IDENT or IDLN_TYPE_ARRAY - check carefully */\n");
+    }
 
     if(IDL_NODE_TYPE(IDL_TYPE_DCL(hinfo->tree).type_spec)
        == IDLN_TYPE_SEQUENCE) {
@@ -343,11 +399,14 @@ cbe_hdr_do_type_dcl(CBEHeaderInfo *hinfo)
       id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_LIST(curitem).data),
 				   "_", 0);
       fprintf(hinfo->of, " %s;\n", id);
-      fprintf(hinfo->of, "extern CORBA_TypeCode _tc_%s;\n", id);
+      fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+      fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", id);
+      fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", id);
 
       switch(IDL_NODE_TYPE(IDL_TYPE_DCL(hinfo->tree).type_spec)) {
 
       case IDLN_TYPE_SEQUENCE:
+	fprintf(hinfo->of, "extern ");
 	orbit_cbe_write_typespec(hinfo->of, IDL_TYPE_SEQUENCE(IDL_TYPE_DCL(hinfo->tree).type_spec).simple_type_spec);
 	fprintf(hinfo->of,
 		"* CORBA_sequence_");
@@ -355,7 +414,15 @@ cbe_hdr_do_type_dcl(CBEHeaderInfo *hinfo)
 	fprintf(hinfo->of, "_allocbuf(CORBA_unsigned_long len);\n");
 
       default:
-	fprintf(hinfo->of, "%s* %s__alloc(void);\n", id, id);
+	fprintf(hinfo->of, "extern %s* %s__alloc(void);\n", id, id);
+	if(IDL_NODE_TYPE(IDL_LIST(curitem).data) != IDLN_TYPE_ARRAY
+	   && IDL_NODE_TYPE(IDL_TYPE_DCL(hinfo->tree).type_spec) != IDLN_TYPE_SEQUENCE) {
+	  fprintf(hinfo->of, "#define %s__free ", id);
+	  orbit_cbe_write_typename(hinfo->of,
+				   IDL_TYPE_DCL(hinfo->tree).type_spec);
+	  fprintf(hinfo->of, "__free\n");
+	} else
+	  fprintf(hinfo->of, "extern gpointer %s__free(gpointer mem, gpointer dat, CORBA_boolean free_strings); /* ORBit internals */\n ", id);
       }
 
       free(id);
@@ -374,12 +441,15 @@ cbe_hdr_do_const_dcl(CBEHeaderInfo *hinfo)
   char *id;
 
   id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_CONST_DCL(hinfo->tree).ident), "_", 0);
+  fprintf(hinfo->of, "#ifndef %s\n", id);
   fprintf(hinfo->of, "#define %s ", id);
-  free(id);
 
   orbit_cbe_write_const(hinfo->of,
 			IDL_CONST_DCL(hinfo->tree).const_exp);
   fprintf(hinfo->of, "\n");
+  fprintf(hinfo->of, "#endif /* !%s */\n\n", id);
+
+  free(id);
 }
 
 static void
@@ -392,6 +462,8 @@ cbe_hdr_do_except_dcl(CBEHeaderInfo *hinfo)
   id =
     IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_EXCEPT_DCL(hinfo->tree).ident),
 			    "_", 0);
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef struct %s {\n", id);
   
   for(curitem = IDL_EXCEPT_DCL(hinfo->tree).members; curitem;
@@ -401,14 +473,19 @@ cbe_hdr_do_except_dcl(CBEHeaderInfo *hinfo)
     fprintf(hinfo->of, "\n");
   }
   fprintf(hinfo->of, "} %s;\n", id);
-  fprintf(hinfo->of, "extern const CORBA_TypeCode _tc_ex_%s;\n", id);
-  fprintf(hinfo->of, "%s* %s__alloc(void);\n", id, id);
-  fprintf(hinfo->of, "#define ex_%s \"ex_%s\"\n", id, id);
+  fprintf(hinfo->of, "#define ex_%s \"%s\"\n", id,
+	  IDL_IDENT(IDL_EXCEPT_DCL(hinfo->tree).ident).repo_id);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", id);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", id);
+  fprintf(hinfo->of, "extern %s* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "extern gpointer %s__free(gpointer mem, gpointer dat, CORBA_boolean free_strings); /* ORBit internals */\n", id);
+
   free(id);
 }
 
 static void
-cbe_hdr_do_attr_dcl(CBEHeaderInfo *hinfo)
+cbe_hdr_do_attr_dcl_internal(CBEHeaderInfo *hinfo, IDL_tree current_interface, gboolean inherited)
 {
   IDL_tree curop, curitem;
   GString *attrname = g_string_new(NULL);
@@ -456,7 +533,10 @@ cbe_hdr_do_attr_dcl(CBEHeaderInfo *hinfo)
 					      IDL_ident_new(strdup("value"))));
 	    }
 	    
-	    orbit_cbe_hdr_process_piece(&subhinfo);
+	    if(inherited==TRUE)
+	      cbe_hdr_do_inherited_op_dcl(&subhinfo, current_interface);
+	    else
+	      orbit_cbe_hdr_process_piece(&subhinfo);
 
 	    /* Restore the fake link to the original in the namespace */
 	    IDL_GENTREE(IDL_IDENT_TO_NS(IDL_LIST(curitem).data)).data = ns_data_save;
@@ -478,6 +558,19 @@ cbe_hdr_do_attr_dcl(CBEHeaderInfo *hinfo)
   g_string_free(attrname, TRUE);
 }
 
+static void
+cbe_hdr_do_attr_dcl(CBEHeaderInfo *hinfo)
+{
+  cbe_hdr_do_attr_dcl_internal(hinfo, NULL, FALSE);
+}
+
+static void
+cbe_hdr_do_inherited_attr_dcl(CBEHeaderInfo *hinfo, IDL_tree current_interface)
+{
+  if(enable_inherited_ops)
+    cbe_hdr_do_attr_dcl_internal(hinfo, current_interface, TRUE);
+}
+
 static void
 cbe_hdr_do_op_dcl(CBEHeaderInfo *hinfo)
 {
@@ -486,23 +579,68 @@ cbe_hdr_do_op_dcl(CBEHeaderInfo *hinfo)
      cut-and-paste of this routine */
 
   char *id, *id2;
-  IDL_tree curitem;
-  int level;
+  IDL_tree curitem, op;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_OP_DCL(hinfo->tree).ident), "_", 0);
+  curitem = IDL_get_parent_node(hinfo->tree, IDLN_INTERFACE, NULL);
+  g_assert(curitem);
+  id2 = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(curitem).ident), "_", 0);
+
+  fprintf(hinfo->of, "extern ");
 
   orbit_cbe_write_typespec(hinfo->of, IDL_OP_DCL(hinfo->tree).op_type_spec);
   orbit_cbe_param_printptrs(hinfo->of, IDL_OP_DCL(hinfo->tree).op_type_spec, DATA_RETURN);
 
-  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_OP_DCL(hinfo->tree).ident), "_", 0);
+  fprintf(hinfo->of, " %s(%s _obj, ", id, id2);
 
-  curitem = IDL_get_parent_node(hinfo->tree, IDLN_INTERFACE, &level);
+  op = hinfo->tree;
+  for(curitem = IDL_OP_DCL(hinfo->tree).parameter_dcls; curitem; curitem = IDL_LIST(curitem).next) {
+    hinfo->tree = IDL_LIST(curitem).data;
+    orbit_cbe_hdr_process_piece(hinfo);
+  }
 
-  g_assert(curitem);
-  id2 = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(curitem).ident), "_", 0);
-  fprintf(hinfo->of, " %s(%s obj, ", id, id2);
-  free(id); free(id2);
+  if(IDL_OP_DCL(op).context_expr)
+    fprintf(hinfo->of, "CORBA_Context ctx, ");
+
+  fprintf(hinfo->of, "CORBA_Environment *ev);\n");
+
+  free(id);
+  free(id2);
+}
+
+/* FIXME This function should check for redefined inherited ops, it doesnt
+   currently - RHP */
+static void
+cbe_hdr_do_inherited_op_dcl(CBEHeaderInfo *hinfo, IDL_tree current_interface)
+{
+  char *id, *id2;
+  IDL_tree ident, curitem, intf;
+
+  if(!enable_inherited_ops)
+    return;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(current_interface).ident), "_", 0);
+  intf = IDL_get_parent_node(hinfo->tree, IDLN_INTERFACE, NULL);
+  id2 = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(intf).ident),
+				"_", 0);
+
+  fprintf(hinfo->of, "extern ");
+
+  orbit_cbe_write_typespec(hinfo->of, IDL_OP_DCL(hinfo->tree).op_type_spec);
+  orbit_cbe_param_printptrs(hinfo->of, IDL_OP_DCL(hinfo->tree).op_type_spec, DATA_RETURN);
+
+  ident = IDL_OP_DCL(hinfo->tree).ident;
+  g_assert(ident);
+
+  fprintf(hinfo->of, " %s_%s(%s _obj, ",
+	  id, IDL_IDENT(ident).str, id);
+  free(id);
+  free(id2);
 
   for(curitem = IDL_OP_DCL(hinfo->tree).parameter_dcls; curitem; curitem = IDL_LIST(curitem).next) {
     hinfo->tree = IDL_LIST(curitem).data;
+
+    /* I dont think anything below here needs to know about inheritance - RHP */
     orbit_cbe_hdr_process_piece(hinfo);
   }
 
@@ -525,7 +663,7 @@ cbe_hdr_do_param_dcl(CBEHeaderInfo *hinfo)
   case IDL_PARAM_INOUT: r = DATA_INOUT; break;
   case IDL_PARAM_OUT: r = DATA_OUT; break;
   default:
-    g_error("Unknown IDL_PARAM type\n");
+    g_error("Unknown IDL_PARAM type");
   }
 
   orbit_cbe_param_printptrs(hinfo->of,
@@ -536,7 +674,22 @@ cbe_hdr_do_param_dcl(CBEHeaderInfo *hinfo)
 static void
 cbe_hdr_do_forward_dcl(CBEHeaderInfo *hinfo)
 {
-  g_assert(!"Not yet implemented");
+  char *id;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(hinfo->tree).ident), "_", 0);
+
+  fprintf(hinfo->of, "/* Begin forward declaration %s */\n\n", id);
+
+  fprintf(hinfo->of, "  /* client-side data structures and methods */\n");
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", id, id);
+  fprintf(hinfo->of, "typedef CORBA_Object %s;\n", id);
+  fprintf(hinfo->of, "#define %s__free CORBA_Object__free\n", id);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+
+  fprintf(hinfo->of, "/* End forward declaration %s */\n\n", id);
+
+  free(id);
 }
 
 static void
@@ -629,12 +782,17 @@ cbe_hdr_do_type_object(CBEHeaderInfo *hinfo)
   fprintf(hinfo->of, "CORBA_Object");
 }
 
+static void
+cbe_hdr_do_type_typecode(CBEHeaderInfo *hinfo)
+{
+  fprintf(hinfo->of, "CORBA_TypeCode");
+}
+
 static void
 cbe_hdr_do_type_enum(CBEHeaderInfo *hinfo)
 {
   IDL_tree curitem;
-  int counter;
-  char *id;
+  char *id, *enumid;
 
   /* CORBA spec says to do
 	   typedef unsigned int enum_name;
@@ -642,21 +800,29 @@ cbe_hdr_do_type_enum(CBEHeaderInfo *hinfo)
      This works just as well and seems cleaner.
   */
 
+  enumid = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_ENUM(hinfo->tree).ident), "_", 0);
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", enumid, enumid);
   fprintf(hinfo->of, "typedef enum {\n");
 
-  for(counter = 0, curitem = IDL_TYPE_ENUM(hinfo->tree).enumerator_list;
+  for(curitem = IDL_TYPE_ENUM(hinfo->tree).enumerator_list;
       curitem;
-      curitem = IDL_LIST(curitem).next, counter++) {
+      curitem = IDL_LIST(curitem).next) {
     id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_LIST(curitem).data), "_", 0);
-    fprintf(hinfo->of, "  %s = %d%s\n", id, counter,
+
+    fprintf(hinfo->of, "  %s%s\n",
+	    id,
 	    IDL_LIST(curitem).next?",":"");
+
     free(id);
   }
 
-  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_ENUM(hinfo->tree).ident), "_", 0);
-  fprintf(hinfo->of, "} %s;\n", id);
-  fprintf(hinfo->of, "extern CORBA_TypeCode _tc_%s;\n", id);
-  free(id);
+  fprintf(hinfo->of, "} %s;\n", enumid);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", enumid);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", enumid);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", enumid);
+
+  free(enumid);
 }
 
 static void
@@ -689,9 +855,14 @@ cbe_hdr_do_type_array(CBEHeaderInfo *hinfo)
   fprintf(hinfo->of, ";\n");
   fprintf(hinfo->of, "typedef %s_slice %s[%qd];\n", id, id,
 	  IDL_INTEGER(IDL_LIST(IDL_TYPE_ARRAY(hinfo->tree).size_list).data).value);
-  fprintf(hinfo->of, "extern CORBA_TypeCode _tc_%s;\n", id);
-  fprintf(hinfo->of, "%s_slice* %s__alloc(void);\n", id, id);
-  fprintf(hinfo->of, "void %s__alloc(%s_slice* ptr);\n", id, id);
+
+  /* The ifdef corresponding to this was written in cbe_hdr_do_type_dcl */
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", id);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", id);
+  fprintf(hinfo->of, "extern %s_slice* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "extern gpointer %s__free(gpointer mem, gpointer dat, CORBA_boolean free_strings); /* ORBit internals */\n", id);
+
   free(id);
 }
 
@@ -703,7 +874,9 @@ cbe_hdr_do_type_struct(CBEHeaderInfo *hinfo)
   CBEHeaderInfo child_hinfo = *hinfo;
 
   id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_STRUCT(hinfo->tree).ident), "_", 0);
-  fprintf(hinfo->of, "struct %s {\n", id);
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", id, id);
+  fprintf(hinfo->of, "typedef struct %s {\n", id);
 
   for(curitem = IDL_TYPE_STRUCT(hinfo->tree).member_list; curitem;
       curitem = IDL_LIST(curitem).next) {
@@ -712,9 +885,12 @@ cbe_hdr_do_type_struct(CBEHeaderInfo *hinfo)
     fprintf(hinfo->of, "\n");
   }
 
-  fprintf(hinfo->of, "};\n");
-  fprintf(hinfo->of, "extern CORBA_TypeCode _tc_%s;\n", id);
-  fprintf(hinfo->of, "struct %s* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "} %s;\n", id);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", id);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", id);
+  fprintf(hinfo->of, "extern %s* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "extern gpointer %s__free(gpointer mem, gpointer dat, CORBA_boolean free_strings);\n", id);
   free(id);
 }
 
@@ -733,6 +909,10 @@ cbe_hdr_do_type_union(CBEHeaderInfo *hinfo)
   IDL_tree curitem;
   CBEHeaderInfo child_hinfo = *hinfo;
 
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_UNION(hinfo->tree).ident), "_", 0);
+
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef struct {\n  ");
   orbit_cbe_write_typespec(hinfo->of, IDL_TYPE_UNION(hinfo->tree).switch_type_spec);
 
@@ -744,10 +924,13 @@ cbe_hdr_do_type_union(CBEHeaderInfo *hinfo)
     orbit_cbe_hdr_process_piece(&child_hinfo);
   }
 
-  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_TYPE_UNION(hinfo->tree).ident), "_", 0);
   fprintf(hinfo->of, "  } _u;\n} %s;\n", id);
-  fprintf(hinfo->of, "extern CORBA_TypeCode _tc_%s;\n", id);
-  fprintf(hinfo->of, "%s* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n", id);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n", id);
+  fprintf(hinfo->of, "extern %s* %s__alloc(void);\n", id, id);
+  fprintf(hinfo->of, "extern gpointer %s__free(gpointer mem, gpointer dat, CORBA_boolean free_strings); /* ORBit internals */\n", id);
+
   free(id);
 }
 
@@ -755,16 +938,29 @@ cbe_hdr_do_type_union(CBEHeaderInfo *hinfo)
 static void
 cbe_hdr_do_member(CBEHeaderInfo *hinfo)
 {
-  IDL_tree curitem;
+  IDL_tree curitem, li, sz;
   fprintf(hinfo->of, "  ");
-  orbit_cbe_write_typespec(hinfo->of, IDL_MEMBER(hinfo->tree).type_spec);
 
   for(curitem = IDL_MEMBER(hinfo->tree).dcls; curitem; curitem = IDL_LIST(curitem).next) {
-    fprintf(hinfo->of, " %s%s",
-	    IDL_IDENT(IDL_LIST(curitem).data).str,
-	    IDL_LIST(curitem).next?", ":"");
+    orbit_cbe_write_typespec(hinfo->of, IDL_MEMBER(hinfo->tree).type_spec);
+    li = IDL_LIST(curitem).data;
+    switch(IDL_NODE_TYPE(li)) {
+    case IDLN_TYPE_ARRAY:
+      fprintf(hinfo->of, " %s",
+	      IDL_IDENT(IDL_TYPE_ARRAY(li).ident).str);
+      for(sz = IDL_TYPE_ARRAY(li).size_list; sz; sz = IDL_LIST(sz).next) {
+	fprintf(hinfo->of, "[%" IDL_LL "d]", IDL_INTEGER(IDL_LIST(sz).data).value);
+      }
+      fprintf(hinfo->of, ";\n");
+      break;
+    case IDLN_IDENT:
+      fprintf(hinfo->of, " %s;\n",
+	      IDL_IDENT(li).str);
+      break;
+    default:
+      g_error("We were asked to print a member for a %s", IDL_tree_type_names[IDL_NODE_TYPE(li)]);
+    }
   }
-  fprintf(hinfo->of, ";");
 }
 
 static void
@@ -785,6 +981,8 @@ cbe_hdr_do_case_stmt(CBEHeaderInfo *hinfo)
       orbit_cbe_write_const(hinfo->of, IDL_LIST(curitem).data);
     else
       fprintf(hinfo->of, "default");
+    if (IDL_LIST(curitem).next)
+      fprintf(hinfo->of, ", ");
   }
   fprintf(hinfo->of, " */\n");
 }
@@ -794,6 +992,7 @@ cbe_hdr_interface_print_epv_for_op(CBEHeaderInfo *hinfo)
 {
   /* hacked from cbe_hdr_do_op_dcl() */
   IDL_tree curitem;
+  
 
   fprintf(hinfo->of, "  ");
   orbit_cbe_write_typespec(hinfo->of, IDL_OP_DCL(hinfo->tree).op_type_spec);
@@ -802,12 +1001,14 @@ cbe_hdr_interface_print_epv_for_op(CBEHeaderInfo *hinfo)
   fprintf(hinfo->of, " (*%s)(PortableServer_Servant servant, ",
 	  IDL_IDENT(IDL_OP_DCL(hinfo->tree).ident).str);
 
+  /* op = hinfo->tree; */
   for(curitem = IDL_OP_DCL(hinfo->tree).parameter_dcls; curitem; curitem = IDL_LIST(curitem).next) {
     hinfo->tree = IDL_LIST(curitem).data;
     orbit_cbe_hdr_process_piece(hinfo);
   }
 
 #if 0
+  /* hinfo->tree.. see op saved above here... */
   /* I'm unsure of how CORBA_Context should be used here... */
   if(IDL_OP_DCL(hinfo->tree).context_expr)
     fprintf(hinfo->of, "CORBA_Context ctx, ");
@@ -883,6 +1084,45 @@ cbe_hdr_interface_print_vepv(IDL_tree node, FILE *of)
   free(id);
 }
 
+static void
+cbe_hdr_interface_print_inheritance_ops(IDL_tree node, CBEHeaderInfo *current)
+{
+  CBEHeaderInfo subhinfo = *current;
+  IDL_tree curitem;
+  char *id;
+
+  g_assert(current->tree);
+  g_assert(IDL_NODE_TYPE(node) == IDLN_INTERFACE);
+
+  if(node==current->tree) {
+    /* We don't inherit from ourself... */
+    return;
+  }
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(node).ident), "_", 0);
+  fprintf(current->of, "/* From interface %s */\n", id);
+  free(id);
+
+  /* Print the operations defined for this interface, but in current's
+     namespace */
+
+  for(curitem = IDL_INTERFACE(node).body;
+      curitem; curitem = IDL_LIST(curitem).next) {
+
+    subhinfo.tree = IDL_LIST(curitem).data;
+
+    switch(IDL_NODE_TYPE(IDL_LIST(curitem).data)) {
+    case IDLN_OP_DCL:
+      cbe_hdr_do_inherited_op_dcl(&subhinfo, current->tree);
+      break;
+    case IDLN_ATTR_DCL:
+      cbe_hdr_do_inherited_attr_dcl(&subhinfo, current->tree);
+      break;
+    default:
+    }
+  }
+}
+
 static void
 cbe_hdr_do_interface(CBEHeaderInfo *hinfo)
 {
@@ -894,32 +1134,58 @@ cbe_hdr_do_interface(CBEHeaderInfo *hinfo)
   fprintf(hinfo->of, "/* Begin interface %s */\n\n", id);
 
   fprintf(hinfo->of, "  /* client-side data structures and methods */\n");
+  fprintf(hinfo->of, "#ifndef _%s_defined\n"
+		     "#define _%s_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef CORBA_Object %s;\n", id);
+  fprintf(hinfo->of, "#define %s__free CORBA_Object__free\n", id);
+  fprintf(hinfo->of, "#endif /* !_%s_defined */\n\n", id);
+  fprintf(hinfo->of, "extern CORBA_unsigned_long %s__classid;\n", id);
+  fprintf(hinfo->of, "extern const struct CORBA_TypeCode_struct TC_%s_struct;\n",
+	  id);
+  fprintf(hinfo->of, "extern const CORBA_TypeCode TC_%s;\n",
+	  id);
   subhinfo.tree = IDL_INTERFACE(hinfo->tree).body;
   cbe_hdr_do_list(&subhinfo);
 
+  if(IDL_INTERFACE(hinfo->tree).inheritance_spec != NULL) {
+    fprintf(hinfo->of, "/* Begin inherited operations for %s */\n\n", id);
+    subhinfo.tree = IDL_INTERFACE(hinfo->tree).inheritance_spec;
+    IDL_tree_traverse_parents(subhinfo.tree, (gpointer)cbe_hdr_interface_print_inheritance_ops, hinfo);
+    fprintf(hinfo->of, "/* End inherited operations for %s */\n\n", id);
+  }
+
   fprintf(hinfo->of, "  /* server-side (via the POA) data structures and methods */\n");
+  fprintf(hinfo->of, "#ifndef _POA_%s__epv_defined\n"
+		     "#define _POA_%s__epv_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef struct {\n");
   fprintf(hinfo->of, "  void *_private;\n");
   cbe_hdr_interface_print_epv(hinfo);
   fprintf(hinfo->of, "} POA_%s__epv;\n", id);
+  fprintf(hinfo->of, "#endif /* !_POA_%s__epv_defined */\n\n", id);
 
+  fprintf(hinfo->of, "#ifndef _POA_%s__vepv_defined\n"
+		     "#define _POA_%s__vepv_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef struct {\n");
   fprintf(hinfo->of, "  PortableServer_ServantBase__epv* _base_epv;\n");
   IDL_tree_traverse_parents(hinfo->tree, (gpointer)cbe_hdr_interface_print_vepv, hinfo->of);
   fprintf(hinfo->of, "} POA_%s__vepv;\n", id);
+  fprintf(hinfo->of, "#endif /* !_POA_%s__vepv_defined */\n\n", id);
 
+  fprintf(hinfo->of, "#ifndef _POA_%s_defined\n"
+		     "#define _POA_%s_defined\n\n", id, id);
   fprintf(hinfo->of, "typedef struct {\n");
   fprintf(hinfo->of, "  void * _private;\n");
   fprintf(hinfo->of, "  POA_%s__vepv* vepv;\n", id);
   fprintf(hinfo->of, "} POA_%s;\n", id);
+  fprintf(hinfo->of, "#endif /* !_POA_%s_defined */\n\n", id);
 
   fprintf(hinfo->of,
-	  "void POA_%s__init(POA_%s *servant, CORBA_Environment *ev);\n", id, id);
+	  "extern void POA_%s__init(POA_%s *servant, CORBA_Environment *ev);\n", id, id);
   fprintf(hinfo->of,
-	  "void POA_%s__fini(POA_%s *servant, CORBA_Environment *ev);\n", id, id);
+	  "extern void POA_%s__fini(POA_%s *servant, CORBA_Environment *ev);\n", id, id);
 
   fprintf(hinfo->of, "/* End interface %s */\n", id);
+    
   free(id);
 }
 
@@ -935,3 +1201,14 @@ cbe_hdr_do_unaryop(CBEHeaderInfo *hinfo)
   g_assert(!"Not yet implemented");
 }
 
+static void
+cbe_hdr_do_native(CBEHeaderInfo *hinfo)
+{
+  char *id;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_NATIVE(hinfo->tree).ident),
+			       "_", 0);  
+  fprintf(hinfo->of, "typedef void *%s;\n", id);
+
+  free(id);
+}
