@@ -20,15 +20,24 @@
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
  *  Author: Dick Porter <dick@cymru.net>
+ *          Elliot Lee <sopwith@redhat.com>
  *
  */
 
+#include <unistd.h>
+#include <errno.h>
+#include <sys/stat.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
-
+#include <pwd.h>
+#include <time.h>
+#include <sys/types.h>
+#include <dirent.h>
+#include "config.h"
+#include "../IIOP/iiop-endianP.h"
 #include "orbit.h"
 
 #include "orbit_poa.h"
@@ -37,24 +46,18 @@
 
 static void ORBit_ORB_release(CORBA_ORB orb, CORBA_Environment *ev);
 
-ORBit_RootObject_Interface CORBA_ORB_epv =
+static const ORBit_RootObject_Interface CORBA_ORB_epv =
 {
-	(gpointer)ORBit_ORB_release,
+	(gpointer)ORBit_ORB_release
 };
 
 
-static int ORBit_ORBid_setup(CORBA_ORB orb, CORBA_ORBid id, CORBA_boolean use_boa)
+static int ORBit_ORBid_setup(CORBA_ORB orb, CORBA_ORBid id)
 {
 	g_assert(orb!=NULL);
 	g_assert(id!=NULL);
 
-	if(!strcmp(id, "orbit-local-orb")) {
-		/* Default to POA */
-		orb->use_poa=use_boa?CORBA_FALSE:CORBA_TRUE;
-	} else if(!strcmp(id, "mico-local-orb")) {
-		/* Fall back to BOA for MICO compatibility */
-		orb->use_poa=CORBA_FALSE;
-	} else {
+	if(strcmp(id, "orbit-local-orb")) {
 		fprintf(stderr, "ORBit_ORBid_setup: Unknown ORB id: %s\n", id);
 		return(0);
 	}
@@ -64,6 +67,65 @@ static int ORBit_ORBid_setup(CORBA_ORB orb, CORBA_ORBid id, CORBA_boolean use_bo
 	return(1);
 }
 
+static void ORBit_rc_load(GString *rcfile, ORBit_orb_options *options)
+{
+	FILE *fp;
+	GHashTable *read_options;
+	ORBit_orb_options *search;
+
+	fp=fopen(rcfile->str, "r");
+	if(fp==NULL) {
+		return;
+	}
+
+	read_options=g_hash_table_new(g_str_hash, g_str_equal);
+
+	while(!feof(fp)) {
+		char buf[1024];
+
+		if(fgets(buf, 1024, fp)!=NULL) {
+			char *bptr, *tmp, *key, *data;
+			size_t start, len;
+
+			if(buf[0]=='#')
+				continue;
+			
+			bptr=buf;
+			tmp=strpbrk(bptr, " \t\n=");
+			if(tmp==NULL) continue;
+			*tmp++='\0';
+			key=g_strdup(bptr);
+			bptr=tmp;
+
+			start=0;
+			while(bptr+start != '\0' &&
+			     (isspace(bptr[start]) || bptr[start]=='='))
+				start++;
+			len=strcspn(bptr+start, " \t\n");
+			bptr[start+len]='\0';
+			if(len>0) {
+				data=g_strdup(bptr+start);
+			} else {
+				data=g_strdup("TRUE");
+			}
+
+			g_hash_table_insert(read_options, key, data);
+		}
+	}
+	fclose(fp);
+
+	for(search=options;search->name!=NULL;search++) {
+		char *read_val;
+
+		read_val=g_hash_table_lookup(read_options, search->name);
+		if(read_val!=NULL) {
+			ORBit_option_set(search, read_val);
+		}
+	}
+
+	g_hash_table_destroy(read_options);
+}
+
 static void ORBit_ORB_release(CORBA_ORB orb, CORBA_Environment *ev) 
 {
 	g_assert(orb!=NULL);
@@ -86,38 +148,117 @@ static void ORBit_ORB_release(CORBA_ORB orb, CORBA_Environment *ev)
 	if(!CORBA_Object_is_nil(orb->root_poa, ev)) {
 		CORBA_Object_release(orb->root_poa, ev);
 	}
-	giop_connection_unref(orb->connection);
+	if(orb->cnx.ipv4)
+		giop_connection_unref(orb->cnx.ipv4);
+	if(orb->cnx.ipv6)
+		giop_connection_unref(orb->cnx.ipv6);
+	if(orb->cnx.usock)
+		giop_connection_unref(orb->cnx.usock);
 
 	g_free(orb);
 }
 
+static GIOPConnection*
+ORBit_ORB_make_usock_connection(void)
+{
+	GIOPConnection *retval = NULL;
+	GString *tmpstr;
+	DIR *dirh;
+	struct dirent *dent;
+	struct stat statbuf;
+
+	tmpstr = g_string_new(NULL);
+
+	g_string_sprintf(tmpstr, "/tmp/orbit-%s", g_get_user_name());
+		
+	if(mkdir(tmpstr->str, 0700) != 0) {
+		int e = errno;
+			
+		switch (e) {
+		case EEXIST:
+			if (stat(tmpstr->str, &statbuf) != 0)
+				g_error ("Can not stat %s\n", tmpstr->str);
+
+			if (statbuf.st_uid != getuid ())
+				g_error ("Owner of %s is not the current user\n",
+					 tmpstr->str);
+
+			if ((statbuf.st_mode & ~(S_IRWXU|S_IFDIR)) != 0)
+				g_error ("Wrong permissions for %s\n",
+					 tmpstr->str);
+			break;
+				
+		default:
+			g_error("Unknown error on directory creation of %s (%s)\n",
+				tmpstr->str, g_strerror (e));
+		}
+	}
+
+	g_string_sprintf(tmpstr, "/tmp/orbit-%s",
+			 g_get_user_name());
+	dirh = opendir(tmpstr->str);
+	while(!retval && (dent = readdir(dirh))) {
+		if(strncmp(dent->d_name, "orb-", 4))
+			continue;
+
+		g_string_sprintf(tmpstr, "/tmp/orbit-%s/%s",
+				 g_get_user_name(), dent->d_name);
+		retval =
+			GIOP_CONNECTION(iiop_connection_server_unix(tmpstr->str));
+	}
+	closedir(dirh);
+
+	srand(time(NULL));
+	while(!retval) {
+		g_string_sprintf(tmpstr, "/tmp/orbit-%s/orb-%d%d",
+				 g_get_user_name(), rand(), rand());
+		retval =
+			GIOP_CONNECTION(iiop_connection_server_unix(tmpstr->str));
+	}
+
+	g_string_free(tmpstr, TRUE);
+
+	return retval;
+}
 
 /* Section 4.4
  *
  * Adjusts argc and argv appropriately
  */
+static CORBA_ORB orb=NULL;
+
 CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, CORBA_Environment *ev)
 {
-	CORBA_ORB orb=NULL;
 	int no_iiop_server=0;
 	int no_iiop_proxy=0;
-	int use_boa=0;
+	int use_ipv4=1;
+	int use_ipv6=0;
+	int use_usock=1;
 	int debug_level=0;
 	int debug_modules=0;
+	int nosysrc=0;
+	int nouserrc=0;
 	char *imr_ior=NULL, *imr_addr=NULL;
 	char *ir_ior=NULL, *ir_addr=NULL;
 	char *naming_ior=NULL, *naming_addr=NULL;
 	char *root_poa_ior=NULL, *root_poa_addr=NULL;
 	char *orb_id_opt=NULL;
+	GString *tmpstr;
 
-	/* The variable addresses in this struct need to be assigned at
+	/* The variable addresses in these structs need to be assigned at
 	 * run-time if you want to compile with -pedantic
 	 *
 	 * (You will also get scads of warnings about "long long" too)
 	 */
 
+	ORBit_orb_options pre_rc_options[]={
+		{"ORBNoSystemRC", no_arg, &nosysrc},
+		{"ORBNoUserRC", no_arg, &nouserrc},
+		{NULL, 0, NULL},
+	};
+
 	/* These options are compatible with MICO */
-	orb_options options[]={
+	ORBit_orb_options options[]={
 		{"ORBNoIIOPServer", no_arg, &no_iiop_server},
 		{"ORBNoIIOPProxy", no_arg, &no_iiop_proxy},
 		{"ORBid", string_arg, &orb_id_opt},
@@ -134,16 +275,32 @@ CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, COR
 
 	/* These options aren't */
 		{"ORBDebugModules", int_arg, &debug_modules},
-		{"ORBUseBOA", no_arg, &use_boa},
 		{"ORBRootPOAIOR", string_arg, &root_poa_ior},
-		{"ORBRootPOAADDR", string_arg, &root_poa_addr},
+		{"ORBRootPOAAddr", string_arg, &root_poa_addr},
+		{"ORBIIOPUSock", int_arg, &use_usock},
+		{"ORBIIOPIPv4", int_arg, &use_ipv4},
+		{"ORBIIOPIPv6", int_arg, &use_ipv6},
 		{NULL,0,NULL},
 	};
 
-	g_return_val_if_fail(ev != NULL, NULL);
-	g_return_val_if_fail(argc != NULL, NULL);
-	g_return_val_if_fail(argv != NULL, NULL);
-	g_return_val_if_fail(orb_identifier != NULL, NULL);
+	g_return_val_if_fail(ev != NULL, orb);
+	g_return_val_if_fail(argc != NULL, orb);
+	g_return_val_if_fail(argv != NULL, orb);
+
+	ORBit_option_parse(argc, argv, pre_rc_options);
+
+	if(!nosysrc) {
+		tmpstr=g_string_new(ORBit_SYSRC);
+		ORBit_rc_load(tmpstr, options);
+		g_string_free(tmpstr, TRUE);
+	}
+
+	if(!nouserrc) {
+		tmpstr = g_string_new(NULL);
+		g_string_sprintf(tmpstr, "%s/.orbitrc", g_get_home_dir());
+		ORBit_rc_load(tmpstr, options);
+		g_string_free(tmpstr, TRUE);
+	}
 
 	ORBit_option_parse(argc, argv, options);
 
@@ -152,7 +309,6 @@ CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, COR
 
 	CORBA_exception_init(ev);
 
-
 	ORBit_chunks_init();
 
 	giop_init();
@@ -164,22 +320,22 @@ CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, COR
 		goto error;
 	}
 
-	ORBit_pseudo_object_init(ORBIT_PSEUDO_OBJECT(orb), ORBIT_PSEUDO_ORB);
+	ORBit_pseudo_object_init(ORBIT_PSEUDO_OBJECT(orb), ORBIT_PSEUDO_ORB, ev);
 	
-	ORBit_RootObject_set_interface(ORBIT_ROOT_OBJECT(orb),&CORBA_ORB_epv,ev);
+	ORBit_RootObject_set_interface(ORBIT_ROOT_OBJECT(orb),
+				       (gpointer)&CORBA_ORB_epv, ev);
 
 	ORBIT_ROOT_OBJECT(orb)->refs = 1;
 
 	if(orb_id_opt!=NULL) {
-		if(!ORBit_ORBid_setup(orb, orb_id_opt, use_boa))
+		if(!ORBit_ORBid_setup(orb, orb_id_opt))
 			goto error;
 		g_free(orb_id_opt);
 	} else if(orb_identifier!=NULL) {
-		if(!ORBit_ORBid_setup(orb, orb_identifier, use_boa))
+		if(!ORBit_ORBid_setup(orb, orb_identifier))
 			goto error;
 	} else {
 		orb->orb_identifier=g_strdup("orbit-local-orb");
-		orb->use_poa=use_boa?CORBA_FALSE:CORBA_TRUE;
 	}
 
 	if(orb->orb_identifier==NULL) {
@@ -187,9 +343,28 @@ CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, COR
 		goto error;
 	}
 
-	orb->connection = GIOP_CONNECTION(iiop_connection_server());
-	giop_connection_ref(orb->connection);
-	GIOP_CONNECTION(orb->connection)->orb_data = orb;
+	if(use_ipv4) {
+		orb->cnx.ipv4 = GIOP_CONNECTION(iiop_connection_server());
+		giop_connection_ref(orb->cnx.ipv4);
+		GIOP_CONNECTION(orb->cnx.ipv4)->orb_data = orb;
+	}
+	if(use_ipv6) {
+		orb->cnx.ipv6 = GIOP_CONNECTION(iiop_connection_server_ipv6());
+		giop_connection_ref(orb->cnx.ipv6);
+		GIOP_CONNECTION(orb->cnx.ipv6)->orb_data = orb;
+	}
+
+	if(use_usock) {
+
+		orb->cnx.usock = ORBit_ORB_make_usock_connection();
+
+		giop_connection_ref(orb->cnx.usock);
+		GIOP_CONNECTION(orb->cnx.usock)->orb_data = orb;
+	}
+
+	orb->objrefs = g_hash_table_new((GHashFunc)g_CORBA_Object_hash,
+					(GCompareFunc)g_CORBA_Object_equal);
+	orb->poas = g_ptr_array_new();
 
 	/* when I figure out what MICO is doing with the iiop_proxy and
 	 * iiop_server stuff, it'll get handled here.
@@ -262,38 +437,29 @@ CORBA_ORB CORBA_ORB_init(int *argc, char **argv, CORBA_ORBid orb_identifier, COR
 	 */
 
 	{
-		PortableServer_POA root_poa=NULL;
-		PortableServer_POAManager poa_mgr=NULL;
-		CORBA_PolicyList policies;
+		PortableServer_POA root_poa=CORBA_OBJECT_NIL;
 	  
 		if(root_poa_ior!=NULL) {
-			root_poa=(PortableServer_POA )CORBA_ORB_string_to_object(orb, root_poa_ior, ev);
+			root_poa=(PortableServer_POA)
+				CORBA_ORB_string_to_object(orb,
+							   root_poa_ior, ev);
 			g_free(root_poa_ior);
-		} else if(root_poa_addr!=NULL) {
-			g_free(root_poa_addr);
-		} else {
-			/* Create a poa manager */
-			poa_mgr = ORBit_POAManager_new();
-
-			/* Initialise the policy list */
-			policies._maximum=0;
-			policies._length=0;
-			policies._buffer=NULL; /* default policies */
-			CORBA_sequence_set_release(&policies,CORBA_TRUE); 
-
-			/* Create the root poa */
-			root_poa = ORBit_POA_new("RootPOA",poa_mgr,&policies,ev);
 		}
 		       
 		/* And attatch it to the orb */
 
 		if(!CORBA_Object_is_nil((CORBA_Object)root_poa, ev)) {
-			CORBA_ORB_set_initial_reference((CORBA_ORB)orb, "RootPOA", (CORBA_Object)root_poa, ev);
+			CORBA_ORB_set_initial_reference((CORBA_ORB)orb,
+							"RootPOA",
+							(CORBA_Object)root_poa,
+							ev);
 		}
 	}
 
 
-	return(orb);
+	ORBit_custom_run_setup(orb, ev);
+
+	return (CORBA_ORB)CORBA_Object_duplicate((CORBA_Object)orb, ev);
 
 error:
 	if(orb!=NULL) {
@@ -329,12 +495,99 @@ error:
 	return(NULL);
 }
 
-CORBA_char *CORBA_ORB_object_to_string(CORBA_ORB orb, CORBA_Object obj, CORBA_Environment *ev)
+typedef struct {
+	CORBA_Object obj;
+	CDR_Codec *codec;
+	gboolean emit_active;
+} profile_user_data;
+
+static void ORBit_emit_profile(gpointer item, gpointer userdata)
+{
+	ORBit_Object_info *profile=(ORBit_Object_info *)item;
+	profile_user_data *data=(profile_user_data *)userdata;
+	CORBA_Object obj=data->obj;
+	CDR_Codec *codec=data->codec, *encaps;
+	gboolean emit_active=data->emit_active;
+	static const CORBA_octet iiopversion[] = {1,0};
+
+	g_assert(obj!=NULL);
+	g_assert(codec!=NULL);
+	g_assert(profile!=NULL);
+
+	if((profile == obj->active_profile) && (emit_active == FALSE))
+		return;			/* we already did this one */
+
+	switch(profile->profile_type) {
+	case IOP_TAG_INTERNET_IOP:
+		encaps = CDR_codec_init();
+		encaps->buffer = g_new(CORBA_octet, 2048);
+		encaps->buf_len = 2048;
+		encaps->readonly = CORBA_FALSE;
+		encaps->host_endian = encaps->data_endian = FLAG_ENDIANNESS;
+
+		CDR_put_ulong(codec, IOP_TAG_INTERNET_IOP);
+		CDR_put_octet(encaps, FLAG_ENDIANNESS);
+		CDR_put_octets(encaps, (gpointer)iiopversion, sizeof(iiopversion));
+		CDR_put_string(encaps, profile->tag.iopinfo.host);
+		CDR_put_ushort(encaps, profile->tag.iopinfo.port);
+		CDR_put_ulong(encaps, profile->object_key._length);
+		CDR_put_octets(encaps, profile->object_key._buffer,
+			profile->object_key._length);
+		CDR_put_ulong(codec, encaps->wptr);
+		CDR_put_octets(codec, encaps->buffer, encaps->wptr);
+		CDR_codec_free(encaps);
+		break;
+
+	case IOP_TAG_ORBIT_SPECIFIC:
+		encaps = CDR_codec_init();
+		encaps->buffer = g_new(CORBA_octet, 2048);
+		encaps->buf_len = 2048;
+		encaps->readonly = CORBA_FALSE;
+		encaps->host_endian = encaps->data_endian = FLAG_ENDIANNESS;
+
+		CDR_put_ulong(codec, IOP_TAG_ORBIT_SPECIFIC);
+		CDR_put_octet(encaps, FLAG_ENDIANNESS);
+		CDR_put_octets(encaps, (gpointer)iiopversion, sizeof(iiopversion));
+		CDR_put_string(encaps, profile->tag.orbitinfo.unix_sock_path);
+		CDR_put_ushort(encaps, profile->tag.orbitinfo.ipv6_port);
+		CDR_put_ulong(encaps, profile->object_key._length);
+		CDR_put_octets(encaps, profile->object_key._buffer,
+			profile->object_key._length);
+		CDR_put_ulong(codec, encaps->wptr);
+		CDR_put_octets(codec, encaps->buffer, encaps->wptr);
+		CDR_codec_free(encaps);
+		break;
+
+	default:
+		g_warning("Skipping tag %d", profile->profile_type);
+		break;
+	}
+}
+
+CORBA_char *CORBA_ORB_object_to_string(CORBA_ORB orb,
+				       CORBA_Object obj,
+				       CORBA_Environment *ev)
 {
   int i;
-  CDR_Codec *codec, *encaps;
+  CDR_Codec *codec;
   GString *retval;
-  CORBA_char *rc;
+  CORBA_char *rc = NULL;
+  CORBA_unsigned_long ntags;
+  profile_user_data data;
+
+  if(!obj || !orb) {
+	  CORBA_exception_set_system(ev,
+				     ex_CORBA_BAD_PARAM,
+				     CORBA_COMPLETED_NO);
+	  return NULL;
+  }
+
+  if(ORBIT_ROOT_OBJECT(obj)->is_pseudo_object) {
+	  CORBA_exception_set_system(ev,
+				     ex_CORBA_MARSHAL,
+				     CORBA_COMPLETED_NO);
+	  return NULL;
+  }
 
   codec = CDR_codec_init();
 
@@ -343,31 +596,20 @@ CORBA_char *CORBA_ORB_object_to_string(CORBA_ORB orb, CORBA_Object obj, CORBA_En
   codec->readonly = CORBA_FALSE;
   codec->host_endian = codec->data_endian = FLAG_ENDIANNESS;
 
-  encaps = CDR_codec_init();
-  encaps->buffer = g_new(CORBA_octet, 2048);
-  encaps->buf_len = 2048;
-  encaps->readonly = CORBA_FALSE;
-  encaps->host_endian = encaps->data_endian = FLAG_ENDIANNESS;
-
   CDR_put_octet(codec, FLAG_ENDIANNESS);
 
-  CDR_put_string(codec, obj->obj_id);
-  CDR_put_ulong(codec, 1); /* Just a TAG_INTERNET_IOP */
-  CDR_put_ulong(codec, IOP_TAG_INTERNET_IOP);
+  CDR_put_string(codec, obj->object_id);
+  ntags = g_slist_length(obj->profile_list);
+  CDR_put_ulong(codec, ntags);
 
-  CDR_put_octet(encaps, FLAG_ENDIANNESS);
+  data.obj=obj;
+  data.codec=codec;
+  data.emit_active=TRUE;
+  if(obj->active_profile != NULL)
+	  ORBit_emit_profile(obj->active_profile, &data); /* do this one first */
 
-  CDR_put_octet(encaps, 1); /* iiop_major */
-  CDR_put_octet(encaps, 0); /* iiop_minor */
-  CDR_put_string(encaps, IIOP_CONNECTION(obj->connection)->hostname);
-  CDR_put_ushort(encaps, ntohs(IIOP_CONNECTION(obj->connection)->location.sin_port));
-  CDR_put_ulong(encaps, obj->object_key._length);
-  CDR_put_octets(encaps, obj->object_key._buffer, obj->object_key._length);
-
-  CDR_put_ulong(codec, encaps->wptr);
-  CDR_put_octets(codec, encaps->buffer, encaps->wptr);
-
-  CDR_codec_free(encaps);
+  data.emit_active=FALSE;
+  g_slist_foreach(obj->profile_list, ORBit_emit_profile, &data);
 
   retval = g_string_new("IOR:");
   for(i = 0; i < codec->wptr; i++)
@@ -375,8 +617,7 @@ CORBA_char *CORBA_ORB_object_to_string(CORBA_ORB orb, CORBA_Object obj, CORBA_En
       g_string_sprintfa(retval, "%02x", codec->buffer[i]);
     }
 
-  rc = CORBA_string_alloc(retval->len);
-  strcpy(rc, retval->str);
+  rc = CORBA_string_dup(retval->str);
   g_string_free(retval, TRUE);
 
   CDR_codec_free(codec);
@@ -384,298 +625,176 @@ CORBA_char *CORBA_ORB_object_to_string(CORBA_ORB orb, CORBA_Object obj, CORBA_En
   return rc;
 }
 
-static void ORBit_ORB_skip_encaps(CDR_Codec *codec)
+CORBA_Object CORBA_ORB_string_to_object(CORBA_ORB orb, CORBA_char *str,
+					CORBA_Environment *ev)
 {
-	CORBA_octet *dump_buffer;
-	CORBA_unsigned_long buf_len;
-
-	if(CDR_get_ulong(codec, &buf_len)==CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "ORBit_ORB_skip_encaps: error reading encapsulation length\n");
-		return;
-	}
-
-	dump_buffer=g_new(CORBA_octet, buf_len);
-	if(dump_buffer==NULL) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "ORBit_ORB_skip_encaps: malloc error creating encapsulation dump_buffer\n");
-		return;
-	}
-
-	/* Copying the buffer this way positions the read
-	 * pointer in "codec" correctly, and gets the
-	 * encapsulation out of our way in case of error.
-	 */
-	if(CDR_buffer_gets(codec, dump_buffer, buf_len)==CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "ORBit_ORB_skip_encaps: error reading encapsulation dump_buffer\n");
-	}
-
-	/* just dump it */
-	g_free(dump_buffer);
-}
-
-CORBA_Object CORBA_ORB_string_to_object(CORBA_ORB orb, CORBA_char *str, CORBA_Environment *ev)
-{
-	CORBA_Object new;
-	CDR_Codec *codec;
-	CORBA_octet *buffer, byteorder;
-	CORBA_char *oid;
-	CORBA_unsigned_long seq_len;
-	CORBA_unsigned_long buf_len;
-	unsigned long len;
-	int i,j;
-
-	g_return_val_if_fail(ev != NULL, CORBA_OBJECT_NIL);
-	g_return_val_if_fail(orb != NULL, CORBA_OBJECT_NIL);
-	g_return_val_if_fail(str != NULL, CORBA_OBJECT_NIL);
+	GSList *profiles=NULL;
+	CORBA_Object retval = NULL;
+	CORBA_char *type_id;
+	ORBit_Object_info *object_info;
+	CDR_Codec *codec, *encaps_codec = NULL;
+	CORBA_octet *buffer, endian;
+	int i, j;
+	CORBA_unsigned_long len, seq_len, misclen;
 
 	if(strncmp(str, "IOR:", 4)) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: unknown IOR reference\n");
+		CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
+					   CORBA_COMPLETED_NO);
 		return(CORBA_OBJECT_NIL);
 	}
 
-	codec=CDR_codec_init();
-	if(codec==NULL) {
-		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
+	codec = CDR_codec_init();
+	len = strlen(str);
+
+	if((len % 2) || len <= 4) {
+		CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
+					   CORBA_COMPLETED_NO);
 		return(CORBA_OBJECT_NIL);
 	}
 
-	len=strlen(str);
-	if(len % 2) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: IOR reference has odd number of digits\n");
-		CDR_codec_free(codec);
-		return(CORBA_OBJECT_NIL);
-	}
-	if(len<4) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: IOR reference is too short\n");
-		CDR_codec_free(codec);
-		return(CORBA_OBJECT_NIL);
-	}
+	codec->buf_len = (len-4)/2;
+	buffer = g_new(CORBA_octet, codec->buf_len);
 
-	buf_len=(len-4)/2;
-	buffer=g_new(CORBA_octet, buf_len);
-	if(buffer==NULL) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: malloc error creating buffer\n");
-		CDR_codec_free(codec);
-		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
+	if(!buffer) {
+		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY,
+					   CORBA_COMPLETED_NO);
 		return(CORBA_OBJECT_NIL);
 	}
 
 	codec->buffer=buffer;
-	codec->buf_len=buf_len;
-	codec->readonly=CORBA_TRUE;
+	codec->readonly = TRUE;
 
-	j=0;
-	for(i=4; i<len; i+=2) {
-		buffer[j++]=HEXOCTET(str[i], str[i+1]);
-	}
+	for(j = 0, i = 4; i < len; i+=2) {
+		buffer[j++] = HEXOCTET(str[i], str[i+1]);
+	};
 
-	if(CDR_get_octet(codec, &byteorder)==CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading byteorder octet\n");
-		CDR_codec_free(codec);
-		return(CORBA_OBJECT_NIL);
-	}
+	CDR_get_octet(codec, &endian);
 
-	if(byteorder) {
-		codec->data_endian=LittleEndian;
-	} else {
-		codec->data_endian=BigEndian;
-	}
+	codec->data_endian = endian;
+	codec->host_endian = FLAG_ENDIANNESS;
 
-	/* CORBA_char *type_id */
+	CDR_get_string(codec, &type_id);
 
-	if(CDR_get_string(codec, &oid)==CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading oid string\n");
-		CDR_codec_free(codec);
-		return(CORBA_OBJECT_NIL);
-	}
-	ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: OID: %s\n", oid);
+	CDR_get_seq_begin(codec, &seq_len);
 
-	/* CORBA_sequence_TaggedProfile profiles */
-
-	if(CDR_get_seq_begin(codec, &seq_len)==CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading profile list length\n");
-		g_free(oid);
-		CDR_codec_free(codec);
-		return(CORBA_OBJECT_NIL);
-	}
-	ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: %d profiles\n", seq_len);
-
-	/* Use the first TAG_INTERNET_IOP profile */
-	for(i=0; i<seq_len; i++) {
+	for(i = 0; i < seq_len; i++) {
 		IOP_ProfileId tag;
 
-		if(CDR_get_ulong(codec, &tag)==CORBA_FALSE) {
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading profile tag no.%d\n", i);
-			g_free(oid);
-			CDR_codec_free(codec);
-			return(CORBA_OBJECT_NIL);
-		}
+		object_info=g_new0(ORBit_Object_info, 1);
 
-		if(tag==IOP_TAG_INTERNET_IOP) {
-			CDR_Codec *encaps_codec;
-			CORBA_octet iiop_major, iiop_minor;
-			CORBA_char *host;
-			CORBA_unsigned_short port;
-			CORBA_unsigned_long key_len;
-			CORBA_octet *object_key;
+		CDR_get_ulong(codec, &tag);
+		switch(tag) {
+		case IOP_TAG_INTERNET_IOP:
+			CDR_get_ulong(codec, &misclen);
+			encaps_codec = CDR_codec_init();
+			encaps_codec->buffer = g_new(CORBA_octet, misclen);
+			CDR_buffer_gets(codec, encaps_codec->buffer, misclen);
 
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: profile no.%d type TAG_INTERNET_IOP\n", i);
+			encaps_codec->buf_len = misclen;
+			encaps_codec->readonly = CORBA_TRUE;
+			CDR_get_octet(encaps_codec, &endian);
+			encaps_codec->data_endian = endian;
+			encaps_codec->host_endian = FLAG_ENDIANNESS;
 
-			if(CDR_get_ulong(codec, &buf_len)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading encapsulation length\n");
-				continue;
-			}
-
-			encaps_codec=CDR_codec_init();
-			if(encaps_codec==NULL) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error creating encapsulation codec\n");
-				CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
-				continue;
-			}
-
-			encaps_codec->buffer=g_new(CORBA_octet, buf_len);
-			if(encaps_codec->buffer==NULL) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: malloc error creating encapsulation buffer\n");
-				CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
-				goto iiop_error1;
-			}
-
-			/* Copying the buffer this way positions the read
-			 * pointer in "codec" correctly, and gets the
-			 * encapsulation out of our way in case of error.
-			 */
-			if(CDR_buffer_gets(codec, encaps_codec->buffer, buf_len)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading encapsulation buffer\n");
-				goto iiop_error2;
-			}
-
-			encaps_codec->buf_len=buf_len;
-			encaps_codec->readonly=CORBA_TRUE;
-
-			if(CDR_get_octet(encaps_codec, &byteorder)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading encapsulation byteorder\n");
-				goto iiop_error2;
-			}
-			
-			if(byteorder) {
-				encaps_codec->data_endian=LittleEndian;
-			} else {
-				encaps_codec->data_endian=BigEndian;
-			}
-
-			if(CDR_get_octet(encaps_codec, &iiop_major)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading encapsulation iiop_major\n");
-				goto iiop_error2;
-			}
-
-			if(iiop_major!=1) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: encapsulation iiop_major is incompatible version (%d)\n", iiop_major);
-				goto iiop_error2;
-			}
-
-			if(CDR_get_octet(encaps_codec, &iiop_minor)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading encapsulation iiop_minor\n");
-				goto iiop_error2;
-			}
-
-			if(iiop_minor>1) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: encapsulation iiop_minor is incompatible version (%d)\n", iiop_minor);
-				goto iiop_error2;
-			}
-
-			if(CDR_get_string(encaps_codec, &host)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading iiop host\n");
-				goto iiop_error2;
-			}
-			if(CDR_get_ushort(encaps_codec, &port)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading iiop port\n");
-				goto iiop_error3;
-			}
-
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: host: [%s] port: %d (IIOP v%d.%d)\n", host, port, iiop_major, iiop_minor);
-
-			if(CDR_get_seq_begin(encaps_codec, &key_len)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading iiop key length\n");
-				goto iiop_error3;
-			}
-			object_key=g_new(CORBA_octet, key_len);
-			if(object_key==NULL) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: malloc error creating iiop object key\n");
-				CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
-				goto iiop_error3;
-			}
-
-			if(CDR_buffer_gets(encaps_codec, object_key, key_len)==CORBA_FALSE) {
-				ORBit_Trace(TraceMod_ORB, TraceLevel_Notice, "CORBA_ORB_string_to_object: error reading iiop object key\n");
-				goto iiop_error4;
-			}
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: object_key: [%s]\n", object_key);
-
-			if(iiop_minor==1) {
-				/* read tagged component sequence */
-				g_assert(!"IIOP v1.1 not yet implemented");
-			}
-
-			/* Now turn this data into an object */
-
-			new = ORBit_CORBA_Object_new(ev);
-			new->orb=orb;
-			if(ORBit_parse_inet(new, host, port)) {
-				CORBA_exception_set_system(ev, ex_CORBA_COMM_FAILURE, CORBA_COMPLETED_NO);
-				goto iiop_error4;
-			}
+			object_info->profile_type = IOP_TAG_INTERNET_IOP;
+			CDR_get_octet(encaps_codec, &object_info->iiop_major);
+			if(object_info->iiop_major != 1)
+				goto error_out;
+			CDR_get_octet(encaps_codec, &object_info->iiop_minor);
+			CDR_get_string(encaps_codec, &object_info->tag.iopinfo.host);
+			CDR_get_ushort(encaps_codec, &object_info->tag.iopinfo.port);
+			CDR_get_seq_begin(encaps_codec, &object_info->object_key._length);
+			object_info->object_key._maximum = 0;
 
+			/* The POA gives out ORBit_alloc()d profiles, so we have to too */
+			object_info->object_key._buffer = ORBit_alloc(object_info->object_key._length, NULL, NULL);
+			CDR_buffer_gets(encaps_codec, object_info->object_key._buffer,
+					object_info->object_key._length);
+			ORBit_set_object_key(object_info);
+			profiles=g_slist_append(profiles, object_info);
 			CDR_codec_free(encaps_codec);
-			CDR_codec_free(codec);
-			new->obj_id = oid;
-			ORBit_set_object_key(new, object_key, key_len);
-			new = CORBA_Object_duplicate(new, ev);
-			g_free(host);
-			return new;
+			encaps_codec = NULL;
+			break;
 
-iiop_error4:
-			g_free(object_key);
-iiop_error3:
-			g_free(host);
-iiop_error2:
-			g_free(encaps_codec->buffer);
-iiop_error1:
-			g_free(encaps_codec);
-		} else if(tag==IOP_TAG_MULTIPLE_COMPONENTS) {
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: profile no.%d type TAG_MULTIPLE_COMPONENTS\n", i);
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: Skipping profile\n");
+		case IOP_TAG_MULTIPLE_COMPONENTS:
+			/* Just skip any multiple_components data, for now */
+			CDR_get_ulong(codec, &misclen);
+			encaps_codec = CDR_codec_init();
+			encaps_codec->buf_len = misclen;
+			encaps_codec->buffer = g_new(CORBA_octet, misclen);
+			encaps_codec->readonly = CORBA_TRUE;
+			CDR_buffer_gets(codec, encaps_codec->buffer, misclen);
+			CDR_codec_free(encaps_codec); encaps_codec = NULL;
+			break;
 
-			ORBit_ORB_skip_encaps(codec);
-		} else {
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: unknown profile no.%d type %d\n", i, tag);
-			ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: Skipping profile\n");
+		case IOP_TAG_ORBIT_SPECIFIC:
+			CDR_get_ulong(codec, &misclen);
+			encaps_codec = CDR_codec_init();
 
-			ORBit_ORB_skip_encaps(codec);
+			encaps_codec->buffer = g_new(CORBA_octet, misclen);
+			CDR_buffer_gets(codec, encaps_codec->buffer, misclen);
+
+			encaps_codec->buf_len = misclen;
+			encaps_codec->readonly = CORBA_TRUE;
+
+			CDR_get_octet(encaps_codec, &endian);
+			encaps_codec->data_endian = endian;
+			encaps_codec->host_endian = FLAG_ENDIANNESS;
+
+			object_info->profile_type=IOP_TAG_ORBIT_SPECIFIC;
+			CDR_get_octet(encaps_codec, &object_info->iiop_major);
+			if(object_info->iiop_major != 1)
+				goto error_out;
+			CDR_get_octet(encaps_codec, &object_info->iiop_minor);
+			CDR_get_string(encaps_codec, &object_info->tag.orbitinfo.unix_sock_path);
+			CDR_get_ushort(encaps_codec, &object_info->tag.orbitinfo.ipv6_port);
+			CDR_get_seq_begin(encaps_codec, &object_info->object_key._length);
+			object_info->object_key._maximum = 0;
+
+			/* The POA gives out ORBit_alloc()d profiles, so we have to too */
+			object_info->object_key._buffer = ORBit_alloc(object_info->object_key._length, NULL, NULL);
+			CDR_buffer_gets(encaps_codec, object_info->object_key._buffer,
+					object_info->object_key._length);
+			ORBit_set_object_key(object_info);
+			profiles=g_slist_append(profiles, object_info);
+			CDR_codec_free(encaps_codec); encaps_codec = NULL;
+			break;
+		default:
+			g_warning("Unknown tag 0x%x", tag);
+
+			/* Skip it */
+			CDR_get_ulong(codec, &misclen);
+			encaps_codec = CDR_codec_init();
+			encaps_codec->buf_len = misclen;
+			encaps_codec->buffer = g_new(CORBA_octet, misclen);
+			encaps_codec->readonly = CORBA_TRUE;
+			CDR_buffer_gets(codec, encaps_codec->buffer, misclen);
+			CDR_codec_free(encaps_codec); encaps_codec = NULL;
+			break;
 		}
 	}
 
-	while(CDR_get_octet(codec, &byteorder)!=CORBA_FALSE) {
-		ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: trailing data: 0x%x\n", byteorder);
-	}
+	retval = ORBit_create_object_with_info(profiles, type_id, orb, ev);
 
-	/* Didn't find a TAG_INTERNET_IOP profile */
-	ORBit_Trace(TraceMod_ORB, TraceLevel_Debug, "CORBA_ORB_string_to_object: Didn't find a TAG_INTERNET_IOP profile\n");
-
-	g_free(oid);
+ error_out:
+	g_free(type_id);
+	if(encaps_codec)
+		CDR_codec_free(encaps_codec);
 	CDR_codec_free(codec);
 
-	return(CORBA_OBJECT_NIL);
+	return retval;
 }
 
 /* Section 5.6.1 */
 CORBA_Status CORBA_ORB_get_default_context(CORBA_ORB orb, CORBA_Context *ctx, CORBA_Environment *ev)
 {
-	g_return_val_if_fail(ev != NULL, -1);
+	g_return_if_fail(ev != NULL);
 
 	if(!ctx) {
 		CORBA_exception_set_system(ev,
 					   ex_CORBA_BAD_PARAM,
 					   CORBA_COMPLETED_NO);
-		return -1;
+		return;
 	}
 
 	/* I have no clue what this is supposed to be...
@@ -685,8 +804,6 @@ CORBA_Status CORBA_ORB_get_default_context(CORBA_ORB orb, CORBA_Context *ctx, CO
 	*/
 
 	*ctx = g_strdup("orbit-local-orb");
-
-	return 0;
 }
 
 /* Section 4.1.2 */
@@ -709,7 +826,7 @@ CORBA_ORB_ObjectIdList *CORBA_ORB_list_initial_services(CORBA_ORB orb, CORBA_Env
 
 	g_assert(orb!=NULL);
 
-	list= g_new(CORBA_ORB_ObjectIdList, 1);
+	list= g_new0(CORBA_ORB_ObjectIdList, 1);
 	if(list==NULL) {
 		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
 		return(NULL);
@@ -717,7 +834,6 @@ CORBA_ORB_ObjectIdList *CORBA_ORB_list_initial_services(CORBA_ORB orb, CORBA_Env
 
 	list->_maximum=2;
 	list->_length=2;
-	list->_flags=0;
 	list->_buffer=g_new(CORBA_ORB_ObjectId, 2);
 	if(list->_buffer==NULL) {
 		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
@@ -750,27 +866,33 @@ CORBA_Object CORBA_ORB_resolve_initial_references(CORBA_ORB orb, CORBA_ORB_Objec
 	g_assert(orb!=NULL);
 	g_assert(identifier!=NULL);
 
-	if(!strcmp(identifier, "ImplementationRepository")) {
-		if(!CORBA_Object_is_nil(orb->imr, ev)) {
-			return(CORBA_Object_duplicate(orb->imr, ev));
+	if(!strcmp(identifier, "ImplementationRepository"))
+		return CORBA_Object_duplicate(orb->imr, ev);
+	else if(!strcmp(identifier, "InterfaceRepository"))
+		return CORBA_Object_duplicate(orb->ir, ev);
+	else if(!strcmp(identifier, "NameService"))
+		return CORBA_Object_duplicate(orb->naming, ev);
+	else if(!strcmp(identifier, "RootPOA")) {
+		if(CORBA_Object_is_nil(orb->root_poa, ev)) {
+			CORBA_PolicyList policies = {0,0,NULL,CORBA_FALSE};
+			PortableServer_POAManager poa_mgr;
+			/* Create a poa manager */
+			poa_mgr = ORBit_POAManager_new();
+			poa_mgr->orb = orb;
+
+			/* Create the root poa */
+			orb->root_poa = ORBit_POA_new(orb, "RootPOA",
+						      poa_mgr,
+						      &policies,
+						      ev);
 		}
-	} else if(!strcmp(identifier, "InterfaceRepository")) {
-		if(!CORBA_Object_is_nil(orb->ir, ev)) {
-			return(CORBA_Object_duplicate(orb->ir, ev));
-		}
-	} else if(!strcmp(identifier, "NameService")) {
-		if(!CORBA_Object_is_nil(orb->naming, ev)) {
-			return(CORBA_Object_duplicate(orb->naming, ev));
-		}
-	} else if(!strcmp(identifier, "RootPOA")) {
-		if(!CORBA_Object_is_nil(orb->root_poa, ev)) {
-			return(CORBA_Object_duplicate(orb->root_poa, ev));
-		}
-	} else {
-		/* throw user exception: InvalidName */
-		CORBA_exception_set(ev,CORBA_USER_EXCEPTION,ex_CORBA_ORB_InvalidName,NULL);	
-		goto error;
+
+		return CORBA_Object_duplicate(orb->root_poa, ev);
 	}
+
+	/* throw user exception: InvalidName */
+	CORBA_exception_set(ev,CORBA_USER_EXCEPTION,ex_CORBA_ORB_InvalidName,NULL);	
+	goto error;
 error:
 	return(NULL);
 }
@@ -831,11 +953,21 @@ void CORBA_ORB_perform_work(CORBA_ORB orb, CORBA_Environment *ev)
 }
 
 /* Section 4.9.4 */
-void CORBA_ORB_shutdown(CORBA_ORB orb, CORBA_boolean wait_for_completion, CORBA_Environment *ev)
+void
+CORBA_ORB_shutdown(CORBA_ORB orb,
+		   CORBA_boolean wait_for_completion,
+		   CORBA_Environment *ev)
 {
 	/* XXX implement on a per-ORB basis, and also
 	   handle whatever wait_for_completion means */
 
+	if(orb->cnx.ipv4)
+		giop_connection_unref(orb->cnx.ipv4);
+	if(orb->cnx.ipv6)
+		giop_connection_unref(orb->cnx.ipv6);
+	if(orb->cnx.usock)
+		giop_connection_unref(orb->cnx.usock);
+
 	giop_main_quit();
 }
 
@@ -843,17 +975,18 @@ void CORBA_ORB_shutdown(CORBA_ORB orb, CORBA_boolean wait_for_completion, CORBA_
 /* CORBA_ORB_run is in server.c */
 
 /* Section 4.7 */
-CORBA_PolicyType CORBA_Policy__get_policy_type(CORBA_Policy obj, CORBA_Environment *ev)
+CORBA_PolicyType
+CORBA_Policy__get_policy_type(CORBA_Policy obj, CORBA_Environment *ev)
 {
-	g_assert(!"Not yet implemented");
-	return(0);
+	return obj->policy_type;
 }
 
 /* Section 4.7 */
-CORBA_Policy *CORBA_Policy_copy(CORBA_Policy obj, CORBA_Environment *ev)
+CORBA_Policy CORBA_Policy_copy(CORBA_Policy obj, CORBA_Environment *ev)
 {
-	g_assert(!"Not yet implemented");
-	return(NULL);
+	ORBIT_ROOT_OBJECT_REF(obj);
+
+	return obj;
 }
 
 /* Section 4.7
@@ -862,12 +995,13 @@ CORBA_Policy *CORBA_Policy_copy(CORBA_Policy obj, CORBA_Environment *ev)
  */
 void CORBA_Policy_destroy(CORBA_Policy obj, CORBA_Environment *ev)
 {
-	g_assert(!"Not yet implemented");
-	return;
+	ORBIT_ROOT_OBJECT_UNREF(obj);
+	if(ORBIT_ROOT_OBJECT(obj)->refs <= 0)
+		ORBIT_ROOT_OBJECT_release(obj, ev);
 }
 
 /* Section 4.8.2 */
-CORBA_Policy *CORBA_DomainManager_get_domain_policy(CORBA_DomainManager obj, CORBA_PolicyType policy_type, CORBA_Environment *ev)
+CORBA_Policy CORBA_DomainManager_get_domain_policy(CORBA_DomainManager obj, CORBA_PolicyType policy_type, CORBA_Environment *ev)
 {
 	g_assert(!"Not yet implemented");
 	return(NULL);
